#include <math.h>

#include "command.h"
#include "client.h"
#include "stopwatch.h"
#include "utils.h"
#include "worldmodel.h"
#include "perceptron.h"
#include "log.h"
#include "netif.h"

/*******************    Commmand   *******************************/
void Command::move(float X,float Y){//move command
	x = X;
	y = Y;
	type = CMD_move;
	time = situation.CurrentTime;
}

void Command::dash(float pow){//dash command
	power = pow;
	type = CMD_dash;
	time = situation.CurrentTime;
}

void Command::turn(float Moment){
	angle = Moment;
	type = CMD_turn;
	time = situation.CurrentTime;
}

void Command::kick(float pow,float ang){ //kick command
	power = pow;
	angle = ang;
	type = CMD_kick;
	time = situation.CurrentTime;
}

void Command::seize(float ang){ //seize command
	if ( !ClientParam::goalie ) my_error("Only goalies can catch");
	if ( situation.playmode == PM_Before_Kick_Off ) return;
	
	type=CMD_catch;
	angle=ang;
	time = situation.CurrentTime;
}

void Command::say(char* msg){
	sprintf(StringBuffer::cmdbuf,"(say %s)",msg);
	type = CMD_say;
	time = situation.CurrentTime;
}

void Command::turn_neck(float ang){
	angle = ang;
	pose_limitation(angle, ServerParam::min_neck_moment, ServerParam::max_neck_moment);
	
	type = CMD_turn_neck;
	time = situation.CurrentTime;
}

void Command::change_view(){
	type = CMD_change_view;
	time = situation.CurrentTime;
}

void Command::disconnect(){
	type = CMD_bye;
	time = situation.CurrentTime;
}

void Command::stay(float desire_bdfc){
	type = CMD_stay;
	time = situation.CurrentTime;
	angle = desire_bdfc;
}

void Command::Tackle(float Power)
{
	this->power= Power;
	type = CMD_tackle;
	time = situation.CurrentTime;
}

void Command::PointTo(float dir,float dist)
{
	this->angle = dir;
	this->dist= dist;
	type = CMD_pointto;
	time = situation.CurrentTime;
}

void Command::Attentionto(int No)
{
	this->No=No;
	type = CMD_attentionto;
	time = situation.CurrentTime;
}

void Command::donothing(){
	type = CMD_donothing;
	time = situation.CurrentTime;
}

bool Command::valid(){
	return (bool) (type != CMD_none);
}

bool Command::valid(Time time){
	return (bool) ((type != CMD_none && this->time == time)||
		((type == CMD_move) && (this->time == -1)));
}

/***********************    Actions  **********************************/
//maintain the history of actions
Actions::Actions(){
	latest = 0;
	kicks = 0;
	dashes = 0;
	turns = 0;
	turn_necks = 0;
	says = 0;
	lastcatchtime = -5;
}

/************ send initial message *******/
void Actions::send_init_message(){//send initializing message	
	if (ServerParam::IP_reconnect)
		sprintf(StringBuffer::cmdbuf, "(reconnect %s %d)", MyTeamName, ServerParam::IP_reconnect);
	else if (ClientParam::goalie){
		sprintf(StringBuffer::cmdbuf, "(init %s (version %.2f) (goalie))", MyTeamName, ServerParam::version);
	}
	else
		sprintf(StringBuffer::cmdbuf, "(init %s (version %.2f))", MyTeamName, ServerParam::version);
	Network::send_message(StringBuffer::cmdbuf,Network::sock);	
}

void Actions::init_clang()
{
	sprintf(StringBuffer::cmdbuf, "(clang (ver %d %d))", ClientParam::clang_min_version, ClientParam::clang_max_version);
	Network::send_message(StringBuffer::cmdbuf,Network::sock);
}

/****** see if commands take effects  ******************/
void Actions::ActionsInvalidation(){
	turn_valid = Commanded(CMD_turn, situation.LastCycle());
	dash_valid = Commanded(CMD_dash, situation.LastCycle());
	turn_neck_valid = Commanded(CMD_turn_neck, situation.LastCycle());
	kick_valid = Commanded(CMD_kick, situation.LastCycle());
	say_valid = Commanded(CMD_say, situation.LastCycle());
	if (sensory.SensedInfKnown(situation.CurrentTime)){
		SenseBodyInfo& sensebody = sensory.GetSenseBodyInfo(situation.CurrentTime);
		if (sensebody.turns < turns){
			if  (!situation.ClockStopped){
				my_error("Miss %d turns",turns - sensebody.turns);
				DoLog(LOG_MISS,"%d turns", turns - sensebody.turns);
			}
			turneffects.Data(situation.LastCycle()) = 0;
			turn_valid = false;
		}
		if (sensebody.dashes < dashes){
			if (!situation.ClockStopped){
				my_error("Miss %d dashes",dashes - sensebody.dashes);
				DoLog(LOG_MISS,"%d dashs", dashes - sensebody.dashes);
			}
			dasheffects.Data(situation.LastCycle()) = 0;
			usedpower.Data(situation.LastCycle()) = 0;
			dash_valid = false;
		}
		if (sensebody.kicks < kicks){
			if (!situation.ClockStopped){
				my_error("Miss %d kicks",kicks - sensebody.kicks);
				DoLog(LOG_MISS,"%d kicks",kicks - sensebody.kicks);
			}
			kickeffects.Data(situation.LastCycle()) = 0;
			usedpower.Data(situation.LastCycle()) = 0;
			kick_valid = false;
		}
		if (sensebody.turn_necks < turn_necks){
			if (!situation.ClockStopped){
				my_error("Miss %d turn_necks",turn_necks - sensebody.turn_necks);
				DoLog(LOG_MISS,"%d turn_necks",turn_necks - sensebody.turn_necks);
			}
			turn_neckeffects.Data(situation.LastCycle()) = 0;
			turn_neck_valid = false;
		}
		if (sensebody.says < says){
			say_valid = false;
			my_error("Miss %d says",says - sensebody.says);
			DoLog(LOG_MISS,"%d says",says - sensebody.says);
		}
		turns = sensebody.turns;
		dashes = sensebody.dashes;
		kicks = sensebody.kicks;
		says = sensebody.says;
		turn_necks = sensebody.turn_necks;
	}
}

/***************   commands are sent by this function  **************/

/***************   commands are sent by this function  **************/
void Actions::send_action(){
	if (!commands[latest].valid(situation.CurrentTime)){
		my_error("Is this command valid? CMD_type : %d",commands[latest].type);
		return;
	}
	
	switch (commands[latest].type){
	case CMD_attentionto:
		if (commands[latest].No <=0)
			sprintf(StringBuffer::cmdbuf, "(attentionto off)");
		else if (commands[latest].No <=11)
			sprintf(StringBuffer::cmdbuf, "(attentionto %s %d)",MyTeamName,commands[latest].No);
		else if (commands[latest].No <=22)
			sprintf(StringBuffer::cmdbuf, "(attentionto %s %d)",TheirTeamName,commands[latest].No-11);
		else
			sprintf(StringBuffer::cmdbuf, "(attentionto off)");
		break;
	case CMD_pointto:
		if (commands[latest].dist< 0)
			sprintf(StringBuffer::cmdbuf, "(pointto off)");
		else
			sprintf(StringBuffer::cmdbuf, "(pointto %.2f %.2f)",commands[latest].dist,commands[latest].angle);
		break;
	case CMD_tackle:
		sprintf(StringBuffer::cmdbuf, "(tackle %.2f)",commands[latest].power);
		break;
	case CMD_kick:
		sprintf(StringBuffer::cmdbuf,"(kick %.2f %.2f)", commands[latest].power, commands[latest].angle);
		kicks++;
		kickeffects.Setdata(Polar2Vector(commands[latest].power * Kickrate(), NormalizeAngle(commands[latest].angle + Self.bodyfacing)), situation.CurrentTime);
		usedpower.Setdata((float)fabs(commands[latest].power) * (commands[latest].power > 0 ? 1.0f : ClientParam::negativepower_factor), situation.CurrentTime);
		LogMat(LM_Ass, "%d %d %d(%.2f,%.2f),%.2f ", Ass_ActEffect, CMD_kick, situation.CurrentTime, kickeffects.Data(situation.CurrentTime).x, kickeffects.Data(situation.CurrentTime).y); 
		DoLog(LOG_SENDACTION,"kickeffect (%.2f %.2f)", KickEffect(situation.CurrentTime).x, KickEffect(situation.CurrentTime).y);
		break;
	case CMD_dash: 
		sprintf(StringBuffer::cmdbuf,"(dash %.2f)", commands[latest].power);
		dashes++;
		dasheffects.Setdata(Polar2Vector(commands[latest].power * Dashrate(), Self.bodyfacing), situation.CurrentTime);
		usedpower.Setdata((float)fabs(commands[latest].power) * (commands[latest].power > 0 ? 1.0f : ClientParam::negativepower_factor), situation.CurrentTime);
		LogMat(LM_Ass, "%d %d %d(%.2f,%.2f),%.2f ", Ass_ActEffect, CMD_dash, situation.CurrentTime, dasheffects.Data(situation.CurrentTime).x, dasheffects.Data(situation.CurrentTime).y); 
		DoLog(LOG_SENDACTION, "dasheffect (%.2f %.2f)", DashEffect(situation.CurrentTime).x, DashEffect(situation.CurrentTime).y);
		break;
	case CMD_turn: 
		sprintf(StringBuffer::cmdbuf,"(turn %.2f)", commands[latest].angle);
		turns++;
		turneffects.Setdata(commands[latest].angle * Turnrate(), situation.CurrentTime);
		LogMat(LM_Ass, "%d %d %d %.2f ", Ass_ActEffect, CMD_turn, situation.CurrentTime, turneffects.Data(situation.CurrentTime)); 
		DoLog(LOG_SENDACTION, "turneffect %.2f", TurnEffect(situation.CurrentTime));
		break;
	case CMD_turn_neck: 
		sprintf(StringBuffer::cmdbuf,"(turn_neck %.2f)", commands[latest].angle);
		turn_necks++;
		turn_neckeffects.Setdata(commands[latest].angle, situation.CurrentTime);
		LogMat(LM_Ass, "%d %d %d %.2f ", Ass_ActEffect, CMD_turn_neck, situation.CurrentTime, turn_neckeffects.Data(situation.CurrentTime)); 
		DoLog(LOG_SENDACTION, "turnneckeffect %.2f", TurnNeckEffect(situation.CurrentTime));
		break;
	case CMD_catch:
		sprintf(StringBuffer::cmdbuf,"(catch %.2f)", commands[latest].angle);
		lastcatchtime = situation.CurrentTime;
		DoLog(LOG_SENDACTION, "catch %.2f", commands[latest].angle);
		break;
	case CMD_move:
		sprintf(StringBuffer::cmdbuf,"(move %.2f %.2f)",commands[latest].x, commands[latest].y);
		DoLog(LOG_SENDACTION, "move");
		break;
	case CMD_change_view:	//obsolete
		sprintf(StringBuffer::cmdbuf, "(change_view %s %s)", width_string, qual_string);
		DoLog(LOG_SENDACTION, "changeview");
		break;
	case CMD_bye:
		sprintf(StringBuffer::cmdbuf, "(bye)");
		DoLog(LOG_SENDACTION, "bye");
		break;
	case CMD_say:
		DoLog(LOG_SENDACTION, "%s", StringBuffer::cmdbuf);
		break;
	default: 
		return;
	}
	Network::send_message(StringBuffer::cmdbuf,Network::sock);
	//DoLog(LOG_SENDACTION, StringBuffer::cmdbuf);
	latest = (latest + 1)% CP_num_stored_CMDs;
}

void Actions::SetKickEffect(float x, float y, float power, Time time){
	kickeffects.Setdata(Vector(x, y), time);
	usedpower.Setdata(power, time);
}
void Actions::SetDashEffect(float x, float y, float power, Time time){
	dasheffects.Setdata(Vector(x,y), time);
}
void Actions::SetTurnEffect(float v, Time time){
	turneffects.Setdata(v, time);
}
void Actions::SetTurnNeckEffect(float v, Time time){
	turn_neckeffects.Setdata(v, time);
}

void Actions::turn(float Moment){
	if (Moment > ServerParam::max_moment){
		//my_error("Should not turn that positive much!");
		Moment = ServerParam::max_moment;
	}
	if (Moment < ServerParam::min_moment){
		//my_error("Should not turn that negative much!");
		Moment = ServerParam::min_moment;
	}
	commands[latest].turn(Moment);
	send_action();
}

void Actions::dash(float power){
	float max, min;
	max = effort_max_power;
	min = effort_min_power;
	
	pose_limitation(power, min, max);
	
	commands[latest].dash(power);
	send_action();
}

void Actions::kick(float power, float angle){
	if (power > ServerParam::max_power){
		//my_error("Should not kick that positive much!");
		power = ServerParam::max_power;
	}
	
	if (power < 0){
		//my_error("Should not kick that negative much!");
		power = 0;
	}
	
	commands[latest].kick(power, NormalizeAngle(angle));
	send_action();
}

void Actions::seize(float angle){
	commands[latest].seize(NormalizeAngle(angle));
	send_action();
}

void Actions::say(char *msg){
	commands[latest].say(msg);
	send_action();
}

void Actions::move(Vector pos){
	move(pos.x, pos.y);
}

void Actions::move(float X,float Y){
	commands[latest].move(X,Y);
	send_action();
}

void Actions::turn_neck(float ang){
	commands[latest].turn_neck(ang);
	send_action();
}
void Actions::Tackle(float Power)
{
	commands[latest].Tackle(Power);
	send_action();
}

void Actions::PointTo(float dist,float dir)
{
	commands[latest].PointTo(dir,dist);
	send_action();
}

void Actions::Attentionto(int No)
{
	if(No == Self.Focus) return;
	commands[latest].Attentionto(No);
	send_action();
}

void Actions::change_view(VIEWWIDTH width,VIEWQUALITY quality){
	switch (quality){
	case VQ_High: sprintf(qual_string,"high"); break;
	case VQ_Low:  sprintf(qual_string,"low");  break;
	}
	switch (width){
	case VW_Narrow: sprintf(width_string,"narrow"); break;
	case VW_Normal: sprintf(width_string,"normal"); break;
	case VW_Wide:   sprintf(width_string,"wide");   break;
	} 
	sensory.NextViewWidth = width; 
	commands[latest].change_view();
	send_action();
}

void Actions::disconnect(){
	commands[latest].disconnect();
	send_action();
}

void Actions::execute(Command& command){
	switch(command.type){
	case CMD_turn:
		turn(command.angle);
		break;
	case CMD_dash:
		dash(command.power);
		break;
	case CMD_kick:
		kick(command.power, command.angle);
		break;
	case CMD_turn_neck:
		turn_neck(command.angle);
		break;
	case CMD_catch:
		seize(command.angle);
		break;
	case CMD_bye:
		disconnect();
		break;
	case CMD_none:
	default:
		break;
	}
}

bool Actions::Commanded(CMDType type, Time time){
	int i;
	for(i = latest - CP_num_commands_at_one_cycle; i < latest; i++){
		if(commands[CmdIndex(i)].time == time && commands[CmdIndex(i)].type == type)
			return true;
	}
	return false;
}

bool Actions::Valid(CMDType type){
	switch(type){
	case CMD_turn:
		return turn_valid;
	case CMD_kick:      
		return kick_valid;
	case CMD_dash:      
		return dash_valid;
	case CMD_turn_neck: 
		return turn_neck_valid;
	case CMD_say:       
		return say_valid;
	default:            
		return false;
	}
	return false;
}

AngleDeg Actions::Max_TurnAng(){
	return ServerParam::max_moment * Turnrate();
}

AngleDeg Actions::Min_TurnAng(){
	return ServerParam::min_moment * Turnrate();
}

float Actions::Kickrate(){
	return float( ServerParam::kick_power_rate * (1 - 0.25f * fabs(NormalizeAngle(ball.global_angle - Self.bodyfacing)) /180 - 0.25f * (ball.distance - Self.player_size - ServerParam::ball_size) / Self.kickable_margin) * Self.effort);
}

float Actions::Dashrate(){
	return float( Self.dash_power_rate * Self.effort);
}

float Actions::Turnrate(float speed){
	return float( 1 / (1.0f + Self.inertia_moment * speed));
}

float Actions::Turnrate(){
	return float( 1 / (1.0f + Self.inertia_moment * Self.speed));
}

float Actions::Max_KickEffect(){
	if (ball.kickable())
		return ServerParam::max_power * Kickrate();
	else
		return 0;
}

float Actions::Max_DashEffect(){
	return ServerParam::max_power * Dashrate();
}

bool Actions::cancatch(Time time){
	return (time - lastcatchtime > ServerParam::catch_ban_cycle)&&(situation.playmode != PM_My_Goal_Kick);
}

float Actions::Kickrate(Time time){
	return kickrates.Data(time);
}

float Actions::Turnrate(Time time){
	return turnrates.Data(time);
}

float Actions::Dashrate(Time time){
	return dashrates.Data(time);
}

Vector Actions::KickEffect(Time time){
	return kickeffects.IsDataKnown(time) ? kickeffects.Data(time) : 0;
}

Vector Actions::DashEffect(Time time){
	return dasheffects.IsDataKnown(time) ? dasheffects.Data(time) : 0;
}

AngleDeg Actions::TurnEffect(Time time){
	return turneffects.IsDataKnown(time) ? turneffects.Data(time) : 0;
}

AngleDeg Actions::TurnNeckEffect(Time time){
	return turn_neckeffects.IsDataKnown(time) ? turn_neckeffects.Data(time) : 0;
}

float Actions::UsedPower(Time time){
	return usedpower.IsDataKnown(time) ? usedpower.Data(time) : 0;
}

void Actions::PreProcess(){
	kickrates.Setdata(Kickrate(), situation.CurrentTime);
	turnrates.Setdata(Turnrate(), situation.CurrentTime);
	dashrates.Setdata(Dashrate(), situation.CurrentTime);
	kickeffects.Setdata(0, situation.CurrentTime);
	turneffects.Setdata(0, situation.CurrentTime);
	turn_neckeffects.Setdata(0, situation.CurrentTime);
	dasheffects.Setdata(0, situation.CurrentTime);
	effort_max_power = Max(Min(ServerParam::max_power, Self.stamina - ServerParam::effort_dec_thr * ServerParam::stamina_max -1), 0.0f);
	effort_min_power = -Max(Min((float)fabs(ServerParam::min_power), (Self.stamina - ServerParam::effort_dec_thr * ServerParam::stamina_max -2) /2.0f), 0.0f);
	
	max_power = Min(ServerParam::max_power, Self.stamina);
	min_power = -Min((float)fabs(ServerParam::min_power), Self.stamina / 2.0f);
}

Vector Actions::SummarizeKickeffectPos(Time prev, Time last, float speed_decay){
	Vector result = 0;
	float alpha = Exp(speed_decay, float(last - prev));
	for(Time time = prev; time < last; time ++){
		result += KickEffect(time) * ( 1 - alpha) / (1 - speed_decay);
		alpha /= speed_decay;
	}
	return result;
}

Vector Actions::SummarizeKickeffectVel(Vector ivel, Time prev, Time last, float speed_decay, float max_speed){
	for(Time time = prev; time < last; time ++){
		ivel += KickEffect(time);
		if (ivel.mod() > max_speed)
			ivel = ivel / ivel.mod() * max_speed;
		ivel *= speed_decay;
	}
	return ivel;
}

Vector Actions::SummarizeDasheffectPos(Vector ivel, Time prev, Time last, float speed_decay, float max_speed){
	Vector moved_pos = 0.0f;
	for(Time time = prev; time < last; time ++){
		ivel += DashEffect(time);
		if (ivel.mod() > max_speed){
			ivel = ivel / ivel.mod() * max_speed;
		}
		moved_pos += ivel;
		ivel *= speed_decay;
	}
	return moved_pos;
}

Vector Actions::SummarizeDasheffectVel(Vector ivel, Time prev, Time last, float speed_decay, float max_speed){
	for(Time time = prev; time < last; time ++){		
		ivel += DashEffect(time);
		if (ivel.mod() > max_speed){
			ivel = ivel / ivel.mod() * max_speed;
		}
		ivel *= speed_decay;
	}
	return ivel;
}

AngleDeg Actions::SummarizeTurneffect(Time prev, Time last){
	float sum_ang = 0.0f;
	for(Time time = prev; time < last; time ++){
		sum_ang += TurnEffect(time);
	}
	return NormalizeAngle(sum_ang);
}

AngleDeg Actions::SummarizeTurnneckeffect(Time prev, Time last){
	float sum_ang = 0.0f;
	for(Time time = prev; time < last; time ++){
		sum_ang += TurnNeckEffect(time);
	}
	return NormalizeAngle(sum_ang);
}

float Actions::CalcKickRate(AngleDeg ang, float distance){
	return float(ServerParam::kick_power_rate * (1 - 0.25f * fabs(NormalizeAngle(ang)) /180 
		- 0.25f * (distance - Self.player_size - ServerParam::ball_size) / Self.kickable_area) * Self.effort);
}

void Actions::ResetKickEffect(Time time){
	kickeffects.Setdata(0, time);
}

void Actions::ResetDashEffect(Time time){
	dasheffects.Setdata(0, time);
}

void Actions::ResetTurnEffect(Time time){
	turneffects.Setdata(0, time);
}

void Actions::ResetTurnneckEffect(Time time){
	turn_neckeffects.Setdata(0, time);
}
