#include "interception.h"
#include "strategy.h"
#include "worldmodel.h"
#include "skill.h"
#include "agent.h"
#include "log.h"
/***********************   Interception   *****************************************/
Interception::Interception(){
	setsmartit(false);
}

void Interception::SetITPlayer(Player* p){
	dash_rate = p->dash_power_rate * p->effort_max;
	if(!p->Is_goalie){
		kickarea = ServerParam::ball_size + p->player_size + p->kickable_margin;
		reliable_kickarea = (float)(int(kickarea*10)/10.0f);
	}else{
		kickarea = ServerParam::catch_area_l;
		reliable_kickarea = ServerParam::catch_area_l;
	}
	maxdash_effect = ServerParam::max_power * dash_rate * p->effort_max;
	rundash_power = ServerParam::max_power;
	max_speed = p->max_speed;
}

void Interception::SetSelfIT(){
	dash_rate = Skill::action.Dashrate();
	kickarea = ServerParam::ball_size + Self.player_size + Self.kickable_margin;
	reliable_kickarea = (float)(int(kickarea*10)/10.0f);
	maxdash_effect = Skill::action.Max_DashEffect();
	max_speed = Self.max_speed; 
	
	Command command;
	Agent::motion.run(Self.max_speed, command);
	rundash_power = command.power;
}

void Interception::prepare_intercept_info(const Vector& ballpos, const Vector& ballvel, const Vector& startpoint, const Vector& selfvel, float bdfacing){
	int i;
	ball_pos[0] = ballpos;	ball_vel[0] = ballvel; 
	ball_speed[0] = ball_vel[0].mod();
	ballangle = ballvel.Angle();
	for(i = 1; i <= IT_simulate_cycles; i++){
		ball_pos[i] = ball_pos[i-1] + ball_vel[i-1];
		ball_vel[i] = ball_vel[i-1] * ServerParam::ball_decay;
		ball_speed[i] = ball_speed[i-1] * ServerParam::ball_decay;
	}

	ballcourse.SetValues(ballpos, ballangle);
	Vector boderpoint;
	if(PitchInfo.shrinked_field.IsWithin(ballpos)){
		PitchInfo.shrinked_field.RayIntersection(ballcourse, boderpoint);
	}else{
		PitchInfo.field.RayIntersection(ballcourse, boderpoint);
	}	
	dist_2_bounder = ballpos.dist(boderpoint);

	max_ballmove = Min(dist_2_bounder, (float)MoveFormula(ball_speed[0], log(0.1/ball_speed[0])/log(ServerParam::ball_decay), ServerParam::ball_decay)+0.1f);
	
	player_pos[0] = startpoint;	player_vel[0] = selfvel; 
	for(i = 1; i < 3;i++){
		player_pos[i] = player_pos[i-1] + player_vel[i-1];
		player_vel[i] = player_vel[i-1] * ServerParam::player_decay;
	}
	bodyfacing = bdfacing;
}
//Ϊturn intercept Ľ״
//Return value: ǷҪsmart intercept,(yes/smart
bool Interception::Is_smart_intercept(Vector* it_infs, unsigned char it_state){
	if(!Is_smartit_needed) return false;

	float IT_reserved_opp_cycles = 2.0f;
	if(Self.IsForward() && Self.IsMid()){
		IT_reserved_opp_cycles = 1.0f;
	}
		
	float IT_smart_ball_minvel = 0.9f;

	int idx = get_closest_it_index(it_state);

	//ĳЩ²ܻҪsmart intercept
	if(	it_infs[idx].y >= IT_simulate_cycles
		|| it_infs[idx].y > their_min_cycles - IT_reserved_opp_cycles	//ֱƽ
		|| pow(ball_speed[1], it_infs[idx].y) < IT_smart_ball_minvel	//̫
		|| it_infs[idx].x >= dist_2_bounder - IT_reserved_opp_cycles	//
		|| fabs(ballangle) > 60){										// 
		return false;
	}

	float max_dist;
	//ֵ
	max_dist = Min(max_ballmove, (float)MoveFormula(ball_speed[0], their_min_cycles - IT_reserved_opp_cycles, ServerParam::ball_decay));
	//߽
	max_dist = Min(max_dist, dist_2_bounder - IT_reserved_opp_cycles);
	//ֱʱ, ֱӳǰܽ,
	if(fabs(ballangle) > 30){
		Line l;
		l.LineFromRay(ballcourse); 
		float x = l.get_x(player_pos[1].y);
		if(!ballcourse.InRightDir(Vector(x, player_pos[1].y))) return false;		
		max_dist = Min(max_dist, ballcourse.DistanceFromOrigin(Vector(x, player_pos[1].y)));
	} 
	
	max_dist += 0.001f;
	//max_dist = Min(max_dist, 0.5f*(ball_speed[0] + IT_smart_ball_minvel)*(ball_speed[0] - IT_smart_ball_minvel)/ServerParam::ball_decay);
	if(it_state == 6){
		if(it_infs[2].x > max_dist){
			//ֻһpeakpoint+ITʱ, Һβ
			return false;
		}else{//ֻȡ
			it_state = 4;
			idx = 2;
		}
	}

	//it_infs[idx].x - max_dist
	Index<Vector, IT_simulate_cycles> values;
	int i = int(it_infs[idx].y+0.9f) + 1;
	int min_cycle = i;
	if(i > IT_simulate_cycles) return false;

	float max_turnagl = (float)fabs(Skill::action.Max_TurnAng()); 
	float dist = ball_pos[i].dist(ball_pos[0]);// (float)MoveFormula(ball_speed[0], i, ServerParam::ball_decay);
	Vector tmp;
	while(dist < max_dist && i < IT_simulate_cycles){
		while((it_state & 2) && i-1 >= it_infs[1].y && i-1 <= (int)it_infs[2].y){
			//зֵ
			i ++;
			if(i > IT_simulate_cycles) break;
			dist += ball_speed[i-1];
		}
		if(i > IT_simulate_cycles) break;
		if(dist > max_dist) break;

		//ȷһת, ȵ´ٿǸŵ
		Vector gap = ball_pos[i]-player_pos[1];
		if(fabs(NormalizeAngle(gap.Angle() - bodyfacing)) < max_turnagl 
			|| gap.mod2() < kickarea ){
			tmp.x = (float)i;
			tmp.y = FieldInfo.GetOffensiveSensitivity(ball_pos[i]);
			if(fabs(NormalizeAngle((ball_pos[i] - player_pos[1]).Angle() - bodyfacing)) < 7){
				//ת
				tmp.y += 0.03f;
			}
			//ͷΣս
			if(i == min_cycle)	tmp.y -= 0.02f;
			values.Add(tmp);
		}

		i ++;
		dist += ball_speed[i-1];
	}

	if(!values.IsEmpty()){
		int bestvalue = values.Num_datas() - 1 ;
		for(i = bestvalue - 1; i >= 0; i--){
			if(values.Data(i).y > values.Data(bestvalue).y){
				bestvalue = i;
			}
		}
		best_IT_cycle = (int)values.Data(bestvalue).x;
		return true;
	}else{
		return false;
	}
}

bool Interception::smart_intercept(int bestcycle, InterceptionInfo& inf){	
	inf.IT_point = ball_pos[bestcycle];
	inf.IT_cycles = (float)bestcycle;
	inf.IT_valid = true;
	inf.IT_is_closeball = true;

	Vector gap = ball_pos[bestcycle] - player_pos[1];
	float angle = gap.Angle();
	float dist = gap.mod();
	if(dist < kickarea){
		inf.IT_rapidness = 0;
		if(bestcycle < 2){
			inf.IT_angle = 0;
		}else{
			inf.IT_angle = IT_smartstay_angle;
		}
		return true;
	}
	float allow_angle_diff = Min(ASin(Min(kickarea, 1.0f)/dist), 35.0f);		
	if(fabs(NormalizeAngle(angle-bodyfacing) ) < Max(allow_angle_diff, 7.0f)){
		inf.IT_angle = 360.0f;
		inf.IT_rapidness = Min(dist/bestcycle + 0.2f, 1.0f);
		if (fabs(NormalizeAngle(player_vel[0].Angle() - bodyfacing))<90){
			inf.IT_dashpow = (inf.IT_rapidness - player_vel[0].mod()) / Skill::action.Dashrate();
		}else{
			inf.IT_dashpow = (inf.IT_rapidness +  player_vel[0].mod()) / Skill::action.Dashrate();
		}
	}else{
		//ҳһ, Ƿܹ //wait
		inf.IT_angle = angle;
		inf.IT_rapidness = 1.0f;
	}
	inf.IT_valid = true;
	inf.IT_is_closeball = true;
	return true;
}

//self intercept - deliberate consideration
void Interception::smart_self_intercept(const Vector& ballpos, const Vector& ballvel, 
		const Vector& startpoint, const Vector& selfvel, float bodyfacing, InterceptionInfo& inf){

	Vector it_infs[3], it_inf;
	unsigned char it_state;
//1) preparation step
	prepare_intercept_info(ballpos, ballvel, startpoint, selfvel, bodyfacing);
	playercourse.SetValues(startpoint, bodyfacing);

	float dash_miss =(max_speed - maxdash_effect)/maxdash_effect ;
	dash_miss = Min(dash_miss, 0.9f);
	set_intercept_param(kickarea - dash_miss, max_speed);	//reliable_kick_area

//2) turn - intercept 
	Vector rel_pos = player_pos[1] - ball_pos[1];
	rel_pos = rel_pos.Rotate(- ballangle);
	set_intercept_param(kickarea - dash_miss, max_speed);	//reliable_kick_area
	it_state = GT_speed_utils::calc_interception_solution(rel_pos, ball_speed[1], it_infs);
	DoLog(LOG_IT, "TurnIT %u (%.1f,%.1f)(%.1f,%.1f)(%.1f,%.1f) ", it_state, 
		it_infs[0].x, it_infs[0].y, it_infs[1].x, it_infs[1].y, it_infs[2].x, it_infs[2].y);
/*	
	simulate_one_dash_step(player_pos[0], ball_pos[0], player_vel[0],ball_vel[0],bodyfacing,100);
	simulate_one_dash_step(player_pos[0], ball_pos[0], player_vel[0],ball_vel[0],bodyfacing,0);
	rel_pos = player_pos[0] - ball_pos[0];
	rel_pos = rel_pos.Rotate(- ballangle);
	it_state = GT_speed_utils::calc_interception_solution(rel_pos, ball_speed[2], it_infs);
*/	
//3) smart intercept consideration
	if(Is_smart_intercept(it_infs, it_state)){
		//smart intercept
		if(best_IT_cycle > 1){
			smart_intercept(best_IT_cycle, inf);
			DoLog(LOG_IT, "smart it %d", best_IT_cycle);
			if(inf.IT_angle < 200){//turn it
				float dist = player_pos[1].dist(ball_pos[best_IT_cycle]); 
				if(dist + dash_miss > (best_IT_cycle-1)*max_speed){
					//maybe dangerous 
					InterceptionInfo no_turn_inf;	
					float distball_noturn;
					if(interception_test(ballcourse, playercourse, ball_speed[0], bodyfacing, distball_noturn, no_turn_inf) == IT_test_ok){
						if(no_turn_inf.IT_cycles <= best_IT_cycle && no_turn_inf.IT_cycles > 5){
							inf = no_turn_inf;
							DoLog("no turn fix for smart it");
							return;
						}else{
							//no turn intercept is false because getcloseballinterception is more accurate
						}
					}
				}
			}
			/*
			*/
			return;	
		}
	}

//4) close ball consideration
	Vector it_infs_noturn[3];
	unsigned char it_state_noturn;
	rel_pos = player_pos[0] - ball_pos[0];
	rel_pos = rel_pos.Rotate(- ballangle); 
	set_intercept_param(kickarea, max_speed);	//reliable_kick_area
	it_state_noturn = GT_speed_utils::calc_interception_solution(rel_pos, ball_speed[0], it_infs_noturn);
	int idx = GT_speed_utils::get_closest_it_index(it_state_noturn);
	if(it_infs_noturn[idx].y <= IT_closeball_max_cycles){//may exist close interception
		safecloseit = bool(their_min_cycles > IT_closeball_max_cycles+0.5f);
		adversary_cycles = (int)their_min_cycles;
		if(getcloseballinterceptioninfo(player_pos[0], ball_pos[0], player_vel[0], ball_vel[0], bodyfacing, inf)){
			//patches
			if(inf.IT_cycles == IT_closeball_max_cycles && inf.IT_angle < 200.0f && noturn_cycles > 0){//turn it
				if(noturn_cycles < adversary_cycles){
					inf.IT_angle = 360.0f;
					inf.IT_point = ball_pos[noturn_cycles];
					inf.IT_cycles = (float)noturn_cycles;
					inf.IT_is_closeball = true;
					inf.IT_dashpow = ServerParam::max_power;
					DoLog("no turn close ball fix %d", noturn_cycles);
				}
			}else if(inf.IT_cycles < 2 && fabs(NormalizeAngle(ball.TheirGoalAngle() - bodyfacing)) > 50){
				float gap = (ball_pos[1] - player_pos[1]).mod() ;
				if(gap > ServerParam::ball_size + Self.player_size + 0.1f && gap < reliable_kickarea - 0.1f){
					inf.IT_angle = ball.TheirGoalAngle();
					DoLog(LOG_IT, "close ball, turn fix");
				}
			}
			DoLog(LOG_IT, "close ball it %.2f", getball_dist);
			return;
		}
	}

//5) no turn intercept consideration
	float distball_noturn;
	if (interception_test(ballcourse, playercourse, ball_speed[0], bodyfacing, distball_noturn, inf) == IT_test_ok){
		if(inf.IT_cycles >= IT_closeball_max_cycles && inf.IT_cycles >= it_infs_noturn[idx].y ){			
			int idx = get_closest_it_index(it_state);
			inf.IT_cycles = Min(inf.IT_cycles, float(int(it_infs[idx].y + 0.999f)) + 1.0f);
			DoLog(LOG_IT, "no turn it");
			return;
		}else{
			//no turn intercept is false because getcloseballinterception is more accurate
		}
	}

//6) has to turn intercept now.
	inf.IT_valid = true;
	inf.IT_rapidness = GT_speed_utils::Player_Intercept_Speed();
	inf.IT_is_closeball = false;
	if(it_state & 1){
		it_inf = it_infs[0];
	}else if(it_state & 2){
		it_inf = it_infs[1];
		if(it_inf.y <= 4.0f && (it_state & 4)){
			//dangerous peakpoint
			it_inf = it_infs[2];
		}
	}else{
		it_inf = it_infs[2];
	}
	inf.IT_cycles = float(int(it_inf.y + 0.999f));
	inf.IT_cycles += 1.0f;	

	bool turn_more = false;
	if(their_min_cycles > inf.IT_cycles + 1){
		if(MoveFormula(ball_speed[0], inf.IT_cycles+1, ServerParam::ball_decay) < max_ballmove){
			DoLog(LOG_IT, "turn more");
			turn_more = true;
			inf.IT_cycles ++;
		}
	}

	if(inf.IT_cycles < IT_simulate_cycles){
		inf.IT_point = ball_pos[(int)inf.IT_cycles];
	}else{
		it_inf.x = (float)MoveFormula(ball_speed[0], inf.IT_cycles, ServerParam::ball_decay);
		Vector rel_it_pos = Vector(it_inf.x, 0);
		inf.IT_point = rel_it_pos.Rotate(ballangle) + ball_pos[0];
	}
	inf.IT_angle = (inf.IT_point - startpoint).Angle();
	DoLog(LOG_IT, "normal turn it");
	if(turn_more)	inf.IT_cycles --; 
}

void Interception::SetInterceptionInfo(bool reset){
	int i, j, num_my, num_their, num_total;

	num_my = FieldInfo.Num_MyVisiblePlayers();
	num_their = FieldInfo.Num_TheirVisiblePlayers();
	num_total = num_my + num_their;
	
	Vector ballpos, ballvel;
	if(!reset){
		ballpos = ball.Pos(Self.updated_sight_time);
		ballvel = ball.Vel(Self.updated_sight_time);
	}else{
		ballpos = ball.pos;
		ballvel = ball.global_vel; 
	}
	Time sight_gap = situation.CurrentTime - Self.updated_sight_time; 
	if(sight_gap != 0){
		DoLog(LOG_IT, "sight gap%d", sight_gap);
	}

	Player *p;
	for(i = 0; i < num_my; i ++){
		MyPlayers_Interception[i] = FieldInfo.MyVisiblePlayers[i];
	}	
	for(i =0; i < num_their; i ++){
		TheirPlayers_Interception[i] = FieldInfo.TheirVisiblePlayers[i];
	}
	for(i = 0; i<num_total; i++){
		p = &GetPlayer(FieldInfo.Players_Close2Ball[i]);
		SetITPlayer(p);
		if(reset){
			if(!p->bodyfc_guessed){  
				getinterceptioninfo(ballpos, ballvel, p->pos, p->global_vel, p->bodyfacing, p->IT_inf);
			}else{
				getinterceptioninfo(ballpos, ballvel, p->pos, p->global_vel, p->IT_inf);
			}
			DoLog(LOG_IT, "(r)%d,(%.1f, %.1f),%.1f",p->InsideNO, p->pos.x, p->pos.y,
				p->IT_inf.IT_cycles); 
		}else if(!p->Updated_pos()){
			//ûиλʱڹֵ
			p->IT_inf.IT_cycles = Max(0.0f, p->IT_inf.IT_cycles-1);
			DoLog(LOG_IT, "(e)%d,(%.1f, %.1f),%.1f",p->InsideNO, p->pos.x, p->pos.y,
				p->IT_inf.IT_cycles);
		}else if(Self.updated_sight_time == situation.CurrentTime || p->original_postime == situation.CurrentTime){
			if(!p->bodyfc_guessed){
				getinterceptioninfo(ballpos, ballvel, p->pos, p->global_vel, p->bodyfacing, p->IT_inf);
			}else{
				getinterceptioninfo(ballpos, ballvel, p->pos, p->global_vel, p->IT_inf);
			}
			DoLog(LOG_IT, "(n)%d,(%.1f, %.1f),%.1f",p->InsideNO, p->pos.x, p->pos.y,
				p->IT_inf.IT_cycles); 
		}else{
			if(!p->bodyfc_guessed){
				getinterceptioninfo(ballpos, ballvel, p->Pos(Self.updated_sight_time), 
					p->Vel(Self.updated_sight_time), p->bodyfacing, p->IT_inf);
			}else{
				getinterceptioninfo(ballpos, ballvel, p->Pos(Self.updated_sight_time),
					p->Vel(Self.updated_sight_time), p->IT_inf);
			}
			p->IT_inf.IT_cycles -= sight_gap;
			DoLog(LOG_IT, "(f)%d,(%.1f, %.1f),%.1f",p->InsideNO, p->pos.x, p->pos.y,
			p->IT_inf.IT_cycles); 
		}
	}

/* sort players by their ability to intercept the ball */
	for(i = 0; i < num_my - 1; i++)
		for ( j = i + 1; j < num_my; j ++){
			if (MyPlayer(MyPlayers_Interception[j]).IT_inf.IT_cycles < MyPlayer(MyPlayers_Interception[i]).IT_inf.IT_cycles){
				Swap(MyPlayers_Interception[j], MyPlayers_Interception[i]);
			}
		}
	
	for(i = 0; i < num_their - 1; i++)
		for ( j = i + 1; j < num_their; j ++){
			if (TheirPlayer(TheirPlayers_Interception[j]).IT_inf.IT_cycles < TheirPlayer(TheirPlayers_Interception[i]).IT_inf.IT_cycles){
				Swap(TheirPlayers_Interception[j], TheirPlayers_Interception[i]);
			}
		}

	situation.most_promising_controller = -1;

	if (num_my > 0){
		my_min_cycles = MyPlayer(MyPlayers_Interception[0]).IT_inf.IT_cycles;
	}else{
		my_min_cycles = InfCycles;
	}

	if(num_their > 0){
		their_min_cycles = TheirPlayer(TheirPlayers_Interception[0]).IT_inf.IT_cycles;
	}else{
		their_min_cycles = InfCycles;
	}
	
	if(Self.IsForward() && Self.pos.x > 0.0f){
		setsmartit(true);
	}else{
		setsmartit(false);
	}
	SetSelfIT();
	smart_self_intercept(ball.pos, ball.global_vel, Self.pos, Self.global_vel, Self.bodyfacing, Self.IT_inf);
	
	self_cycles = Self.IT_inf.IT_cycles;
	
	situation.min_IT_cycles = InfCycles;

	DoLog("min_it_cycles, their%.1f, my%.1f, me%.1f", their_min_cycles, my_min_cycles, self_cycles);
	//жϣص, ʱ,ƾж
	float min_cycles = Min(my_min_cycles, Min(their_min_cycles, self_cycles));
	int promising_side;//0-opponent, 1-teammate, 2-self
	if(their_min_cycles == min_cycles){
		if(my_min_cycles > min_cycles && self_cycles > min_cycles){
			promising_side= 0;
		}else{
			float d1,d2,d3;
			d1 = TheirPlayer(TheirPlayers_Interception[0]).pos.dist2(TheirPlayer(TheirPlayers_Interception[0]).IT_inf.IT_point);
			if(my_min_cycles == min_cycles){
				d2 = MyPlayer(MyPlayers_Interception[0]).pos.dist2(MyPlayer(MyPlayers_Interception[0]).IT_inf.IT_point);
			}else{
				d2 = d1+1;
			}
			if(self_cycles = min_cycles){
				d3 = Self.pos.dist2(Self.IT_inf.IT_point);
			}else{
				d3= d1+1;
			}
			if(d1<=d2 && d1<=d3)
				promising_side = 0;
			else if(d2<=d3)
				promising_side = 1;
			else
				promising_side = 2;

		}
	}else if(my_min_cycles== min_cycles){
		if(self_cycles > min_cycles){
			promising_side = 1;
		}else{
			float d1,d2;
			d1 = MyPlayer(MyPlayers_Interception[0]).pos.dist2(MyPlayer(MyPlayers_Interception[0]).IT_inf.IT_point);			
			d2 = Self.pos.dist2(Self.IT_inf.IT_point);
			promising_side = (d1<=d2 ? 1 : 2);			
		}
	}else{
		promising_side = 2;
	}
	if(promising_side == 0){
		if(TheirPlayers_Interception[0] <= SP_team_size){
			situation.most_promising_controller = SP_team_size + TheirPlayers_Interception[0];
		}
		else{
			situation.most_promising_controller = TheirPlayers_Interception[0];
		}

		situation.next_interception_point = TheirPlayer(TheirPlayers_Interception[0]).IT_inf.IT_point;
		situation.min_IT_cycles = TheirPlayer(TheirPlayers_Interception[0]).IT_inf.IT_cycles;
	}else if(promising_side == 1){
		situation.most_promising_controller = MyPlayers_Interception[0];
		situation.next_interception_point = MyPlayer(MyPlayers_Interception[0]).IT_inf.IT_point;
		situation.min_IT_cycles = MyPlayer(MyPlayers_Interception[0]).IT_inf.IT_cycles;
	}else if(promising_side == 2){
		situation.most_promising_controller = Agent::MyNumber;
		situation.next_interception_point = Self.IT_inf.IT_point;
		situation.min_IT_cycles = Self.IT_inf.IT_cycles;
	}
	my_min_cycles = Min(Self.IT_inf.IT_cycles, my_min_cycles);
	situation.Team_Control = calculatecontrol(my_min_cycles, their_min_cycles);
}

IT_mode Interception::intercept(InterceptionInfo& inf){
	Command command;
	IT_mode it_mode = intercept(inf, command);
	Agent::mediator.enroll(command, Action_interception, PriorityA);
	return it_mode;	
}

IT_mode Interception::intercept(InterceptionInfo& inf, Command& command){
	//intercept action
	command.Reset();
	if (inf.IT_is_closeball){
		if (inf.IT_angle == IT_dash_angle){
			command.dash(inf.IT_dashpow);
		}else if(inf.IT_angle == IT_smartstay_angle){
			command.stay(0);
		}else{
			Agent::motion.turn_to(inf.IT_angle, command);
		}
		DoLog(LOG_IT, "closeballinterception");
		return IT_closeball;		
	}
	else{
		if (inf.IT_valid && fabs(NormalizeAngle(inf.IT_angle - Self.bodyfacing)) > 15 - 8*exp(-ball.distance * 0.15f)){
			Agent::motion.turn_to(inf.IT_angle, command);
			return IT_turn;
		}
		else{
			Agent::motion.run(inf.IT_rapidness, command);
			return IT_dash;
		}
	}
}

void Interception::Do_Intercept(InterceptionInfo& IT_info){
	Command command;
	intercept(IT_info, command);	
	Agent::mediator.enroll(command, Action_interception, 2 * PriorityA);
}

bool Interception::SmartInterception(){
	//first decide whether the player should intercept the ball
	//if he is in interception, decide whether he should cancel it

	//the sight information is sufficient to make interception decision
	if (ball.pos_conf < 0.75 || ball.speed_conf < 0.7 || ball.Isforgot()) return false;
	
	//ball is in someone's control, don't be hook up
	if (!situation.BallFree) return false;

	DoLog(LOG_IT, "most promising IT player%d, cycle%.2f at (%.2f %.2f)",situation.most_promising_controller,
		GetPlayer(situation.most_promising_controller).IT_inf.IT_cycles, GetPlayer(situation.most_promising_controller).pos.x, GetPlayer(situation.most_promising_controller).pos.y);
	DoLog(LOG_IT, "My Cycles %.2f", Self.IT_inf.IT_cycles);
	
	// can not get the ball
	if (Self.IT_inf.IT_cycles == InfCycles ) return false;

	float control;
	int num = FieldInfo.Num_MyVisiblePlayers();
	int i;
	Player* p;
	float IT_extra_gap = 2.0f;
//check if there is need for me to intercept that ball
	for(i = 0; i < num; i ++){

		if (MyPlayer_CtrlOfBall(i).Is_goalie 
			|| Agent::MyNumber == MyPlayers_Interception[i]) continue;

		control = calculatecontrol(Self.IT_inf, MyPlayer_CtrlOfBall(i).IT_inf, true);

		//fix by yjy
		if(their_min_cycles <= MyPlayer_CtrlOfBall(i).IT_inf.IT_cycles && Self.IT_inf.IT_cycles < MyPlayer_CtrlOfBall(i).IT_inf.IT_cycles){
			break;
		}else if(their_min_cycles <= Self.IT_inf.IT_cycles && Self.IT_inf.IT_cycles > MyPlayer_CtrlOfBall(i).IT_inf.IT_cycles){
			return false;
		}

		if (control > 0.55f){
		//the player with the best chance of getting the ball should go
			DoLog(LOG_IT, "I have better chance than %d", MyPlayers_Interception[i]);
			break;
		}
		else if (WouldIntercept(situation.MySide, MyPlayers_Interception[i])){
			if (control < 0.45f){
				DoLog(LOG_IT, "%d has a better chance whose cycle is %f", MyPlayers_Interception[i], MyPlayer_CtrlOfBall(i).IT_inf.IT_cycles);
				return false;
			}
			else if (!Is_IT_Priored(Agent::MyNumber, MyPlayers_Interception[i])){
			//is it dangerous
				DoLog(LOG_IT, "%d is equal in chance but it is priored whose cycle is %.2f", MyPlayers_Interception[i], MyPlayer_CtrlOfBall(i).IT_inf.IT_cycles);
				return false;
			}
			else{
				DoLog(LOG_IT, "I am priored than %d", MyPlayers_Interception[i]);
				break;
			}
		}
	}


	bool give_up = false;
	//see if I can get ball from opponent
	num = FieldInfo.Num_TheirVisiblePlayers();
	for(i = 0; i < num; i ++){
		control = calculatecontrol(Self.IT_inf, TheirPlayer_CtrlOfBall(i).IT_inf);
		if(control >= 0.45f){
		//there is great chance
			break;
		}

		if(Self.aggressiveness > 0.5f){
			if (control >= 0.3f)
				break;
		}
		if (WouldIntercept(situation.TheirSide, TheirPlayers_Interception[i])){
				// there is no chance for me to get the ball from opponent
			give_up = true;
			DoLog(LOG_IT, "There is no chance to get ball from %d whose cycles is %.2f", TheirPlayers_Interception[i], TheirPlayer_CtrlOfBall(i).IT_inf.IT_cycles);
			p = &TheirPlayer(TheirPlayers_Interception[i]);
			if(p->Is_goalie){
			}else if(Self.IsForward() && Self.pos.dist(p->IT_inf.IT_point) < p->pos.dist(p->IT_inf.IT_point) + IT_extra_gap){
			}else{
				return false;
			}
		}
	}
	
	//if the ball is definitely out of pitch before i can get there, let it go
	if(!PitchInfo.WithInExpandedField(Self.IT_inf.IT_point)){
		DoLog(LOG_IT, "Out of pitch(%.2f,%.2f)",Self.IT_inf.IT_point.x,  Self.IT_inf.IT_point.y);
		Ray ballray(ball.pos, ball.global_vel.Angle());
		Vector it_point;
		
		if(!PitchInfo.expanded_field.RayIntersection(ballray, it_point)) return false; 

		int t = int(Self.pos.dist(it_point) / Self.IT_inf.IT_rapidness);
		if(ball.global_vel.mod()  * (1 - Exp(ball.speed_decay, (float)t))/(1-ball.speed_decay) < ball.pos.dist(it_point)){
			DoLog("ITPatch:maybe in pitch");
		}else{
			return false;
		}
	}

	//patch for back's dangerous it
	if(Self.IT_inf.IT_cycles > TheirPlayer_CtrlOfBall(i).IT_inf.IT_cycles ){
		if(Self.IsBack()){
			DoLog(LOG_IT, "Dangerous back it");
			if(TheirPlayers_Interception[0] != FieldInfo.TheirPlayers_Close2Me[0]){
				DoLog(LOG_IT, "patch, cancel it for wrong player");
				return false;
			}else{
				Command cmd;
				intercept(Self.IT_inf, cmd);
				if(cmd.GetCMDType() == CMD_turn){
					DoLog(LOG_IT, "patch, cancel it for turn");
					return false;
				}
			}
		}
	}
	if(give_up){
		p = &TheirPlayer(TheirPlayers_Interception[0]);
		if(p->Is_goalie){ 
			DoLog(LOG_IT, "intercept from goalie %.0f-%.0f", p->IT_inf.IT_cycles, Self.IT_inf.IT_cycles);
		}else{
			DoLog(LOG_IT, "intercept from opp%d %.0f-%.0f", TheirPlayers_Interception[0], p->IT_inf.IT_cycles, Self.IT_inf.IT_cycles);
		}
	}
	Do_Intercept(Self.IT_inf);
	return true;
}

bool Interception::Is_IT_Priored(UNum player1, UNum player2){
	if (!MyPlayer(player1).IsRoleKnown()) return false;
	if (!MyPlayer(player2).IsRoleKnown()) return true;
	Vector point1, point2, ballpos;
	point1 = formation.GetFormationPoint(MyPlayer(player1));
	point2 = formation.GetFormationPoint(MyPlayer(player2));

	ballpos = situation.NextControlBallPos();
	return ballpos.dist2(point1) < ballpos.dist2(point2);
}


bool Interception::WouldIntercept(char side, UNum No){
	if (GetPlayer(side, No).Is_goalie){
		return PitchInfo.their_penaltyarea.IsWithin(GetPlayer(side, No).IT_inf.IT_point);
	}
	return true;
}

float Interception::calculatecontrol(InterceptionInfo& myinf, InterceptionInfo& theirinf, bool is_teammate){
	if (myinf.IT_is_closeball && !is_teammate){
		return calculatecontrol(myinf.IT_cycles, (float)int(theirinf.IT_cycles) + 2.0f);
	}
	return calculatecontrol(myinf.IT_cycles, theirinf.IT_cycles);
}

float Interception::calculatecontrol(float my_cycles, float their_cycles){
	if(my_cycles == InfCycles){
		if (their_cycles == InfCycles)
			return 0.5f;
		else
			return 0.0f;
	}
	else if (their_cycles == InfCycles)
		return 1.0f;
	double input[2];
	double control;
	float cycles_gap = (float)fabs(my_cycles - their_cycles), min_cycles = Min(my_cycles, their_cycles);
	input[0] = fabs(my_cycles - their_cycles) /10;
	input[1] = fabs(Min(my_cycles, their_cycles)) / 10;
	ControlBallNet.SimulateNet(&input[0], &control);
	control = Max(control, 0.5);
	
	if (my_cycles <= their_cycles)
		return  (float)control;
	else
		return 1 - (float)control;
}


Player& Interception::MyPlayer_CtrlOfBall(int idx){
	return MyPlayer(MyPlayers_Interception[idx]);
}

Player& Interception::TheirPlayer_CtrlOfBall(int idx){
	DoLog(LOG_SEE,"TheirPlayer_CtrlOfBall(%d) %d",idx,TheirPlayers_Interception[idx]);
	return TheirPlayer(TheirPlayers_Interception[idx]);
}
