#include "objects.h"
#include "worldmodel.h"
#include "perceptron.h"
#include "skill.h"
#include "log.h"

ObjectUpdate::ObjectUpdate(){
	pos_time = -1;
}

void ObjectUpdate::set_polar(AngleDeg ang,Time time){
	//do not exchange the order of the next two statements
	seenposinfo.Setdata(Vector(seenposinfo.Data(seentime).x, ang), time);
	seentime = Max(seentime, time);
	seen = true;
}
  
void ObjectUpdate::set_polar(AngleDeg ang,float d,Time time){
	seenposinfo.Setdata(Vector(d, ang), time);
	seen = true;
	if(seentime < time){
		seentime = time;
		sdtime = time;
		seen_times.Add(time);
	}else if(seentime > time){
		int i,j;
		for(i=seen_times.Num_datas()-2; i>=0; i--){
			if(seen_times.Data(i) == time) return;
			if(seen_times.Data(i) < time){
				break;
			}
		}
		i++;
		seen_times.Add(time);
		for(j=seen_times.Num_datas()-1; j>i; j--){
			seen_times.Data(j) = seen_times.Data(j-1);
		} 
		seen_times.Data(j) = time;
	}else{
		if(seen_times.IsEmpty()){
			seen_times.Add(time);
		} 
	}
}


void ObjectUpdate::set_conf_param(float max,float decay,float min){
	max_conf = max;
	conf_decay = decay;
	min_valid_conf = min;
}

void ObjectUpdate::set_objtype(ObjType objtype){
	this->objtype = objtype;
}

/*******************************   Object   **************************/
Object::Object(){
	pos_conf = 0;
	global_angle = 0;
}

void Object::set_pos(float X,float Y,bool rotate){
	if (!rotate)
		pos = Vector(X,Y);
	else
		pos = Vector(-X,-Y);
}

void Object::set_global_angle(AngleDeg ang,bool rotate){
	if(!rotate)
		global_angle = ang;
	else
		global_angle = NormalizeAngle(180 + ang);
}

bool Object::pos_valid(){
	return bool(pos_conf >= min_valid_conf);
}

/**************************Mobile Object Update Process************************/
MobileUpdate::MobileUpdate(){
	coordinate_time = -1;
	heard_pos = false;
	heard_vel = false;
	pos_delta_valid = false;
}

void MobileUpdate::estimate_pos_delta(Time t){	
	if(!pos_delta_valid) return;

	while(pos_delta_time < t){
		pos_deltas.Setdata(pos_deltas.Data(pos_delta_time) + max_speed, pos_delta_time+1);
		pos_delta_time ++;
	}
}


void MobileUpdate::Parse_HearPos(TimedData<NoiseVector>& hearpos){
	if(!heard_pos){
		heardpos_info = hearpos;
		heard_pos = true;
	}else if(hearpos.data.delta >= hearpos.data.delta){
		//the second hear info is more valuable
		heardpos_info = hearpos;
	}
}

void MobileUpdate::Parse_HearVel(TimedData<NoiseVector>& hearvel){
	if(!heard_vel){
		heardvel_info = hearvel;
		heard_vel = true;
	}else if(hearvel.data.delta >= hearvel.data.delta){
		//the second hear info is more valuable
		heardvel_info = hearvel;
	}
}

void MobileUpdate::set_speed_param(float max,float decay){
	max_speed = max;
	speed_decay = decay;
}

void MobileUpdate::set_chinfo(float distChng,float dirChng,Time time){
	seenchinfo.Setdata(Vector(distChng, dirChng), time);
}

/***************************   Mobile Object       *********************/
MobileObject::MobileObject(){
	Object::Object();
	speed_conf = 0;
	speed_conf = 0;
	global_vel = 0;
	update_seen_time = -1;
	seentime = -1;
	sdtime = -1;
}

Vector MobileObject::Pos(Time time){
	return gpos.IsDataKnown(time) ? gpos.Data(time) : pos;
}

int MobileObject::PosDelay(){
	return situation.CurrentTime - original_postime;
}

int MobileObject::VelDelay(){
	return situation.CurrentTime - original_veltime;
}

Vector MobileObject::Vel(Time time){
	return gvel.IsDataKnown(time) ? gvel.Data(time) : global_vel;
}

float MobileObject::ActiveRadius(Time time){
	return pos_deltas.ChkData(time);
}

void MobileObject::update_end(){
	if(newpos){
		forgot = false;
	}
#ifndef _Up_FState
	if(sensory.NewSight && !seen && inview(sensory.LastSightTime) && distance < ServerParam::team_far_length){
		DoLog(LOG_UPDATE, "forget obj (%.2f %.2f)", pos.x, pos.y);
		forget();
	}
#endif
	pos = Pos(situation.CurrentTime);
	rel_pos = pos - Self.pos;
	global_angle = rel_pos.Angle();
	distance = rel_pos.mod();

	if (domain.IsDataKnown(situation.CurrentTime)) 
	{
		RingSector * rs = &(domain.Data(situation.CurrentTime));
		if (objtype == OBJ_Player && rs->IsValid() &&			//valid player and valid domain
			!rs->IsWithin(global_angle, distance)				//pos is out of domain
			&& rel_pos.mod() < ServerParam::team_far_length * 0.9f){	//too far player needn't care
			/*
			if(ball.kickable()){
				DoLog("ctrller adjust");
			} 
			//player pos correction based on domain, move from outside into the edge of domain
			DoLog("move player %d into domain mean%.1f delta%.1f, former dir%.1f",
				((Player*) this)->InsideNO, 
				rs->angles.Data(0).mean,
				rs->angles.Data(0).delta,
				global_angle);
			*/
			//distance correction
			if (distance > rs->radius.sup()){
				distance = rs->radius.sup();
			}else if (distance < rs->radius.inf()){ 
				distance = rs->radius.inf();
			}
			if(!rs->IsWithin(global_angle)){
				//angle correction, it's assumed to lies in the first sector of domain.
				float da = NormalizeAngle(global_angle - rs->angles.Data(0).mean);
				if (da > rs->angles.Data(0).delta){
					global_angle = NormalizeAngle(rs->angles.Data(0).sup());
				}else if(da < - rs->angles.Data(0).delta){
					global_angle = NormalizeAngle(rs->angles.Data(0).inf());
				}
			}
			rel_pos = Polar2Vector(distance, global_angle);
			pos = Self.pos + rel_pos;
			//DoLog("from (%.1f, %.1f) to (%.1f, %.1f)", Pos(situation.CurrentTime).x, Pos(situation.CurrentTime).y, pos.x, pos.y);
		}
	}
	
	global_vel = Vel(situation.CurrentTime);
	speed = global_vel.mod();
	vel_angle = global_vel.Angle();
//yjy
	pos_delta = pos_deltas.Data(situation.CurrentTime);

	seen = false;
	isupdateheard = false;
	seen_times.Reset();
	heard_pos = false;
	heard_vel = false;
}

//
void MobileObject::Prior_update(){
	newpos = false;
	updated_seen = false;
	newvel = false;
}

void MobileObject::update(Time time){
	update_seen(time);
	update_heard(time);

	//if(IsValidObject()){
	estimate(time);
	pos = Pos(sensory.LastSightTime);
	rel_pos = pos - Self.Pos(sensory.LastSightTime);
	distance = rel_pos.mod();
	//}
}

void MobileObject::update_seen(Time time){
#ifdef	_Up_FState
	if(sensory.HasFullState() && acc_vel.IsDataKnown(time)){
		gvel.Setdata(acc_vel.Data(time), time);
		gpos.Setdata(acc_pos.Data(time), time);
		pos_time = time;
		coordinate_time = time;
		vel_time = time;
		speed_conf = max_conf;
		pos_conf = max_conf;
		update_seen_time = time;
		original_postime = time;
		original_veltime = time;

		updated_seen = true;
		newpos = true;
		newvel = true;

		pos_delta_valid = true;
		pos_deltas.Setdata(0,time);
		pos_delta_time = time;
		rel_pos = gpos.Data(time) - Self.Pos(time);
		domain.Data(time).Set(rel_pos.mod(), 0, rel_pos.Angle(), 0); 
		domain.TimedData(time).time = time;
	} 
#else
//ûµӾϢ, ֱӷ
	if(seen){
//յӾϢ,ʱθ (һ12ӾϢ)
		Vector pos, rel_vel, rpos;
		int i = 0, t = 0;
		for(; i<seen_times.Num_datas(); i++){
			t = seen_times.Data(i);
			/*       Position         */
			rpos = Polar2Vector(seen_distance(t), seen_rel_angle(t)); 
			pos = rpos.Rotate(Self.Headfacing(t)) + Self.Pos(t);
			estimate(t -1);
			gpos.Setdata(pos, t);
			
			/*****************/
			float seegap = Sqr(seen_distance(t)*ServerParam::dist_qstep*0.5f) + Sqr(seen_distance(t)*ServerParam::dir_qstep*0.5f );
			pos_deltas.Setdata((float)sqrt(seegap), t);
			
			/*       Velocity         */
			if (seenchinfo.IsDataKnown(t)){
				newvel = true;
				
				rpos.Normalize();
				rel_vel.x = seen_distCh(t) * rpos.x - (seen_dirCh(t) * PI / 180.0f * seen_distance(t) * rpos.y);
				rel_vel.y = seen_distCh(t) * rpos.y + (seen_dirCh(t) * PI / 180.0f * seen_distance(t) * rpos.x);
				rel_vel = rel_vel.Rotate(Self.Headfacing(t));

				gvel.Setdata(Self.Vel(t) + rel_vel, t);
				vel_time = t;
				speed_conf = max_conf;
				
				original_veltime = t;
			}
		}
		updated_seen = true;
		newpos = true;

		update_seen_time = t;
		pos_time = t;
		coordinate_time = t;
		pos_conf = max_conf;
		original_postime = t;
		pos_delta_time = t;
		pos_delta_valid = true;
		forgot = false;
	}
#endif
}

void MobileObject::estimate(Time time){
	//if(!IsValidObject()) return;

	Time t;
	if(vel_time < pos_time - 5){//for unknown player, maybe no old vel;
		vel_time = pos_time;
		gvel.Setdata(0, vel_time);
		speed_conf = 0;
	}
	for(t = vel_time; t < pos_time; t ++){
		estimate_vel(t + 1);
	}
	if(sensory.LastSightTime > coordinate_time && coordinate_time > 0 && gpos.IsDataKnown(coordinate_time)){
		float posconfbk = pos_conf;
		for(t = coordinate_time; t < pos_time; t++){			
			estimate_pos(t + 1);
		}
		pos_conf = posconfbk;
	}
	for(t = pos_time; t < time; t ++){
		estimate_vel(t + 1);
		estimate_pos(t + 1);
	}
}

void MobileObject::estimate_vel(Time time){
	Vector evel = gvel.ChkData(time -1);

	if (objtype == OBJ_Ball)
		evel += Skill::action.KickEffect(time -1);
	if (evel.mod() > max_speed){
		evel = evel / evel.mod() * max_speed;
	}
	evel *= speed_decay;
	gvel.Setdata(evel, time);
	speed_conf *= conf_decay;
	vel_time = time;
	if (objtype == OBJ_Ball){
		DoLog(LOG_UPDATE, "Estimate Ball Moving with (%.2f %.2f) at %d", Vel(time).x, Vel(time).y, time);
	}
}

void MobileObject::estimate_pos(Time time){
	Vector epos = Pos(time -1) + gvel.ChkData(time) / speed_decay + Self.pos_gap.ChkData(time);	
	gpos.Setdata(epos , time);
	pos_conf *= conf_decay;
	pos_time = time;
	if(Self.pos_gap.IsDataKnown(time)){
		coordinate_time = time;
	}
}

//overloaded function, no use
bool MobileObject::Isupdate_heard(){
	if(!heard_pos) return false;
	if(!pos_valid()) return true;
	return true;
}

void MobileObject::update_heard(Time time){
#ifdef	_Up_FState
	isupdateheard = false;
	return ;
#endif	
	isupdateheard = Isupdate_heard();

	if (isupdateheard){
		newpos = true;	
		gpos.Setdata(heardpos_info.data.mean, heardpos_info.time);
		pos_time = heardpos_info.time;
		coordinate_time = pos_time;

		//0.2f ˫Ķλ. ˺һʱ,ϵԼϵƫ
		float relateive_delta = heardpos_info.data.delta + CP_coordinate_gap_between_agents;
		relateive_delta -= (heardpos_info.data.mean - Self.pos).mod() * ServerParam::dist_qstep; 
		relateive_delta = Max(relateive_delta, 0.0f);

		original_postime = heardpos_info.time;
		pos_conf = (float)pow(conf_decay, relateive_delta);

		//ûкoriginal_postime,ʱȡĸ
		//original_postime = heardpos_info.time;
		
		pos_delta_time = heardpos_info.time;
		pos_deltas.Setdata(heardpos_info.data.delta + CP_coordinate_gap_between_agents, pos_delta_time);
		pos_delta_valid = true;
	}
}

bool MobileObject::inview(Time time){
	if (!IsValidObject()) return false;

	if (pos_time < time){
		estimate(time);
	}
	Vector tmp_pos = (Pos(time) - Self.Pos(time)).Rotate(-Self.Headfacing(time));
	if(tmp_pos.mod() < ServerParam::feel_distance * 0.5f) return true; 
	//DoLog(7,"The Object is assumed to be at rel(%.2f %.2f)", tmp_pos.x, tmp_pos.y);
	float slide_dist = 1.0f / Sin(Self.MyViewAngle(Self.viewwidth.Data(time)));
	tmp_pos -= Vector(slide_dist, 0.0f);
	return bool(fabs(tmp_pos.Angle()) <= Self.MyViewAngle(Self.viewwidth.Data(time)));
}

void MobileObject::forget(){
	forgot = true;
}

bool MobileObject::speed_valid(){
	return bool(speed_conf > min_valid_conf);
}

Vector MobileObject::PredictPos(int cycles){
	Vector predictpos = pos;
	Vector vel = global_vel;
	int i;
	for(i = 0; i < cycles;i++){
		predictpos += vel;
		vel *= speed_decay;
	}
	return predictpos;
}

void MobileObject::PredictPos(Vector p[], int cycles){
	if(cycles <= 0) return;
	Vector vel = global_vel;
	p[0] = pos + vel;
	int i;
	for(i = 1; i < cycles;i++){
		vel *= speed_decay;
		p[i] = p[i-1] + vel;
	}
}

float MobileObject::MinDist(const Vector& tpos){
	float Min_dist = domain.Data(situation.CurrentTime).Dist(tpos - Self.pos);
	Min_dist = Max(Min_dist, pos.dist(tpos) - pos_delta);
	return Min_dist;
}

float MobileObject::Max_posx(){
	float max = pos.x + pos_delta;
	max = Min(max, domain.Data(situation.CurrentTime).Getmax_x() + Self.pos.x);
	return max;
}

float MobileObject::Min_posx(){
	float min = pos.x - pos_delta;
	min = Max(min, domain.Data(situation.CurrentTime).Getmin_x() + Self.pos.x);
	return min;
}

float MobileObject::GetOffsensitivity(){
	return GetOffsensitivity(situation.CurrentTime);
}

float MobileObject::GetOffsensitivity(Time t){
	if(!offsensitivity.IsDataKnown(t)){
		offsensitivity.Setdata(FieldInfo.GetOffensiveSensitivity(Pos(t)), t);
	}
	return offsensitivity.Data(t);
}

float MobileObject::GetDefsensitivity(){
	return GetDefsensitivity(situation.CurrentTime);
}

float MobileObject::GetDefsensitivity(Time t){
	if(!defsensitivity.IsDataKnown(t)){
		defsensitivity.Setdata(FieldInfo.GetDefensiveSensitivity(Pos(t)), t);
	}
	return defsensitivity.Data(t);
}

bool MobileObject::IsLastSeen(){
	return bool(Last_seen_time() == Self.Last_seen_time()) ;
}


/*********************     Ball     **************************/
Ball::Ball(){
	posgap_time = -1;
	velgap_time = -1;
}

void Ball::update_seengap(Time t){
#ifdef _Up_FState
	if(sensory.HasFullState() && acc_vel.IsDataKnown(t)){
		posgap.Setdata(0, t);
		posgap_time = t;
		velgap.Setdata(0, t);
		velgap_time = t;
		return;
	}
#endif

	if(!seenposinfo.IsDataKnown(t)) return;

	Vector t_posgap;
	float headfacingap = 1.5f;
	float dist = seen_distance(t);
	float rposagl = NormalizeAngle(Self.Headfacing(t) +  seen_rel_angle(t));
	float fcosagl = (float)fabs(Cos(rposagl));
	float fsinagl = (float)fabs(Sin(rposagl));
	//ԾǶ
	float dis_gap = 0.05f + 0.513f * dist * ServerParam::dist_qstep/0.9f;
	float agl_gap = Deg2Rad(0.5f + headfacingap);
	//ͶӰȫϵ
	t_posgap.x = dis_gap * fcosagl + (dist + dis_gap) * agl_gap * fsinagl;
	t_posgap.y = dis_gap * fsinagl + (dist + dis_gap) * agl_gap * fcosagl;
	posgap.Setdata(t_posgap,t);
	
	Vector rpos = Pos(t) - Self.Pos(t);
	DoLog(LOG_UPDATE,"ball rpos(%.4f,%.4f) gap(%.4f %.4f) hafcing %.1f at %d", rpos.x,rpos.y, t_posgap.x, t_posgap.y, Self.Headfacing(t),t);
	posgap_time = t;
	if(seenchinfo.IsDataKnown(t)){
		Vector t_velgap;
	
		//ٶ
		t_velgap.x = dis_gap / dist * (float)fabs(seen_distCh(t)) + (dist + dis_gap) * ServerParam::disch_qstep * 0.5f;
		t_velgap.y  = dis_gap * Deg2Rad(seen_dirCh(t)) + (dist + dis_gap) * Deg2Rad(ServerParam::dir_qstep * 0.5f);
		
		t_velgap = t_velgap.Rotate(rposagl);
		t_velgap.x = (float)fabs(t_velgap.x);
		t_velgap.y = (float)fabs(t_velgap.y);
		//תrposagl
		float tmp = t_velgap.y;
		t_velgap.y += t_velgap.x * agl_gap;
		t_velgap.x += tmp * agl_gap;
		
		//Selfٶ
		Vector svelgap;
		float rvelagl = NormalizeAngle(Self.Headfacing(t) + Self.SpeedAngle(t));
		fcosagl = (float)fabs(Cos(rvelagl));
		fsinagl = (float)fabs(Sin(rvelagl));
		svelgap.x = (Self.Speed(t) + 0.005f) * agl_gap * fsinagl;
		svelgap.y = (Self.Speed(t) + 0.005f) * agl_gap * fcosagl;
		t_velgap += svelgap;
		velgap.Setdata(t_velgap,t);
		velgap_time = t;

		DoLog(LOG_UPDATE,"ball vel(%.4f,%.4f) gap(%.4f %.4f) selfvel gap(%.2f,%.2f)",Vel(t).x,Vel(t).y,t_velgap.x,t_velgap.y,svelgap.x,svelgap.y);
	}
}

void Ball::estimategap(Time t){
#ifdef _Up_FState
	if(sensory.HasFullState() && acc_vel.IsDataKnown(t)){  
		velgap.Setdata(0, t);
		velgap_time = t;
		posgap.Setdata(0, t);
		posgap_time = t;
		return;
	}

#endif
	Time i;
	Vector gap;
	for(i = velgap_time; i < t; i++){
		gap = (velgap.ChkData(i) + Vector(1,1) * ServerParam::ball_rand * (velgap.ChkData(i).mod() + gvel.ChkData(i+1).mod()/speed_decay)) * speed_decay;
		gap.x = gap.x > max_speed ? max_speed : gap.x;
		gap.y = gap.y > max_speed ? max_speed : gap.y;
		velgap.Setdata(gap, i+1);
		DoLog(LOG_UPDATE,"Estimate velgap (%.3f,%.3f) at %d", velgap.ChkData(i+1).x, velgap.ChkData(i+1).y, i+1);
	}
	velgap_time = t;
	//DoLog(LOG_UPDATE, "Estimate posgap (%.3f,%.3f) at %d(%.2f %.2f %.2f)",posgap.ChkData(posgap_time).x, posgap.ChkData(posgap_time).y, posgap_time, speed_decay, ServerParam::player_rand, ServerParam::player_decay);
	for(i = posgap_time; i < t; i++){
		posgap.Setdata(posgap.ChkData(i) + velgap.ChkData(i+1)/speed_decay + Vector(1,1) * ServerParam::player_rand * Self.Vel(i+1).mod() / ServerParam::player_decay, i+1);
		//DoLog(LOG_UPDATE, "Estimate posgap (%.3f,%.3f) selfvel (%.2f,%.2f) at %d",posgap.ChkData(i+1).x, posgap.ChkData(i+1).y, Self.Vel(i+1).x, Self.Vel(i+1).y, i+1);
	}
	posgap_time = t;
}

void Ball::update_seen(Time time){
/* What's special about ball's updation?
	Collisions, and a more accurate velocity is needed*/
#ifdef _Up_FState
	if(sensory.HasFullState() && acc_pos.IsDataKnown(time)){
		MobileObject::update_seen(time);
		return;
	}
#endif
	if(!seen) return;
	Time last_update_time = Max(update_seen_time, time -2);	
	Vector predictpos, predictvel;
	Vector allowposgap, allowvelgap;
	Vector predictpos_gap, predictvel_gap;
	
	int t;
	for(t = last_update_time + 1; t <= time; t++){
		if(!seenposinfo.IsDataKnown(t)) continue;	
		estimate(t);
		predictpos = Pos(t) - Self.Pos(t);
		DoLog(LOG_UPDATE, "estimate relpos %.2f, %.2f", predictpos.x, predictpos.y);
		predictvel = Vel(t);
		allowposgap = posgap.ChkData(t);
		allowvelgap = velgap.ChkData(t);
		Time last_seen_time = original_postime;
		Time last_seench_time = original_veltime;
		MobileObject::update_seen(t);
		update_seengap(t);
		
		if(seen_distance(t) > 30.0f) continue;
		
		if (seen_distance(t) > (Self.kickable_area+0.04f)){
			Skill::action.ResetKickEffect(t); //then the kick command sent at seen_time must be a fault
		}
		int postime_gap = t - last_seen_time;
		if(postime_gap <= 0){
			DoLog(LOG_UPDATE,"why postime_gap = 0? %d", t);
			continue;
		}
		
		allowposgap += posgap.ChkData(t);
		predictpos_gap = Pos(t) - Self.Pos(t) - predictpos;// - Self.SummarizePosGap(last_seen_time, t);
		DoLog(LOG_UPDATE,"T gap%d, V gap%d, Predictpos Gap(%.2f,%.2f)", postime_gap, t - last_seench_time, predictpos_gap.x, predictpos_gap.y);
		
		bool selfcoll = Self.SelfColl(last_seen_time+1, t);
		float erroridx = 0;
		bool PredictError = IsposError(allowposgap, predictpos_gap, erroridx);
		
		if(seenchinfo.IsDataKnown(t)){
			allowvelgap += velgap.ChkData(t);
			predictvel_gap = Vel(t) - predictvel;
			DoLog(LOG_UPDATE,"predictvel Gap(%.2f,%.2f)",predictvel_gap.x,predictvel_gap.y);
			float verroridx;
			bool PrevelError = IsposError(allowvelgap, predictvel_gap, verroridx);
			if(PredictError || PrevelError){
				erroridx += verroridx;
				DoLog(LOG_UPDATE, "Error Index %.3f",erroridx);
				find_kick = true;
			}else if(predictvel_gap.mod2() > Sqr(0.3f)){
				find_kick = true;
			} 
			error_idx.Setdata(erroridx, t);
			continue;
		}else if(Self.IsMoved(last_seen_time, t)){
			continue;
		}
		if(predictpos.mod() < ServerParam::ball_size + ServerParam::player_size){
			DoLog(LOG_UPDATE, "predict coll, unhandled yet");
		}
		error_idx.Setdata(erroridx, t);
		float sqrconf = max_conf * Sqr(conf_decay);
		float CP_collision_maxdis = 0.75f;
		
		if(!PredictError ){
			if(postime_gap ==1 && !OtherKick(last_seen_time,t) && velgap.IsDataKnown(t) && (posgap.Data(t).mod() + posgap.Data(t-1).mod()) > velgap.Data(t).mod()){
				DoLog(LOG_UPDATE, "estimate maybe more accurate, reset velconf");
				speed_conf = max_conf;
				//wait for estimate
			}else if(selfcoll && postime_gap <= 2){
				DoLog(LOG_UPDATE, "self coll, pos still correct fortunately");
				//wait for estimate
			}else if(!gpos.IsDataKnown(last_seen_time)){
				DoLog(LOG_UPDATE, "last pos is lost, have to estimate");
				//wait for estimate
			}else if(postime_gap<4 && speed_valid()){
				ApplyPosbased_vel(last_seen_time,t,max_conf);
			}else if(postime_gap<6){
				ApplyPosbased_vel(last_seen_time,t,sqrconf);
			}else
				ApplyPosbased_vel(last_seen_time,t,sqrconf * sqrconf);
			
			if(!selfcoll && postime_gap == 1 && seen_distance(t) < 2.0f){
				original_veltime = t - 1;
				DoLog(LOG_UPDATE,"Reliable predict vel at %d", t);
			}
			continue;
		}else{
			find_kick = true;
			DoLog(LOG_UPDATE,"why ball is here, Error Index %.3f at %d", erroridx, t);
		}

		//unmoved posgap limit;
		float dist_moved;
		if(gpos.IsDataKnown(last_seen_time)){
			dist_moved = (gpos.Data(last_seen_time) + Self.SummarizePosGap(last_seen_time, t)).dist(gpos.Data(t)) - allowposgap.mod();
		}else{
			dist_moved = predictpos.dist(gpos.Data(t)) - Vel(t).mod() * postime_gap  - allowposgap.mod();
		}
		if(dist_moved > max_speed * postime_gap){
			DoLog(LOG_UPDATE,"ball is moved");
			gvel.Setdata(0, t);
			vel_time = t;
			if(postime_gap ==1){
				speed_conf = max_conf;
			}else if(postime_gap == 2){
				speed_conf = sqrconf;
			}
		}else if(postime_gap ==1){
			if(OtherKick(last_seen_time)){
				ApplyPosbased_vel(last_seen_time,t,sqrconf);
				DoLog(LOG_UPDATE,"Predict other kick at %.0f",(float)last_seen_time);
			}else if(selfcoll){
				//wait for estimate
				DoLog(LOG_UPDATE,"Error because Self coll");
			}else if(last_seench_time < last_seen_time-1 ){
				ApplyPosbased_vel(last_seen_time, t, sqrconf );
				DoLog(LOG_UPDATE,"Error maybe too old vel");
			}else if(seen_distance(t) < CP_collision_maxdis){
				Collisionvel(t,sqrconf);
				DoLog(LOG_UPDATE,"It must be coll");
			}else{
				ApplyPosbased_vel(last_seen_time, t, sqrconf);
				DoLog(LOG_UPDATE,"Unknown reason");
			}
		}else if(postime_gap == 2){
			if(OtherKick(last_seen_time)){
				ApplyPosbased_vel(last_seen_time, t, sqrconf * conf_decay);
				DoLog(LOG_UPDATE,"Predict other kick2 at%d", last_seen_time);
			}else if(OtherKick(last_seen_time + 1)){
				ApplyPosbased_vel(last_seen_time + 1, t, sqrconf * conf_decay);
				DoLog(LOG_UPDATE,"Predict other kick3 at%d", last_seen_time+1);
			}else if(selfcoll){
				//wait for estimate
				DoLog(LOG_UPDATE,"Error because Self coll");
			}else if(last_seench_time<last_seen_time-1){
				ApplyPosbased_vel(last_seen_time, t, sqrconf * conf_decay );
				DoLog(LOG_UPDATE,"Error maybe too old vel");
			}else{
				if(seen_distance(t) > ServerParam::feel_distance){
					DoLog(LOG_UPDATE,"too far ,see wrong");
				}else{
					float dis1 = (Pos(t) - Self.Pos(t)).mod() ;
					float dis2 = (Pos(t-1) - Self.Pos(t-1)).mod();
					DoLog(LOG_UPDATE,"t%d dis%.2f, t%d dis%.2f", t-1, dis2, t,dis1);
					if(dis1 < dis2 && dis1 < CP_collision_maxdis){
						Collisionvel(t, sqrconf * conf_decay);
						DoLog(LOG_UPDATE,"find coll at t%d", t);
					}else if(dis2 < dis1 && dis2 < CP_collision_maxdis){
						Collisionvel(t - 1, sqrconf * conf_decay);
						DoLog(LOG_UPDATE,"find coll at t%d", t-1);
					}else{
						if(Skill::action.KickEffect(last_seen_time).mod() > 0 && ball.Pos(last_seen_time).dist2(Self.Pos(last_seen_time)) > 1.0f){
							DoLog(LOG_UPDATE,"maybe kick%d invalid", last_seen_time+1);
							Skill::action.ResetKickEffect(last_seen_time+1);
						}else{
							DoLog(LOG_UPDATE,"maybe unfound kick %d", t);
						}

						ApplyPosbased_vel(last_seen_time, t, sqrconf * conf_decay );
						//invalid kick not be clear out. or other Skill::kick haven't been found
					}
				}
			}
		}else if(!gpos.IsDataKnown(last_seen_time)){
			//wait for estimate
		}else if(postime_gap<6){
			ApplyPosbased_vel(last_seen_time, t, sqrconf * conf_decay);
		}else{
			ApplyPosbased_vel(last_seen_time, t, ClientParam::min_valid_conf * conf_decay);
		}
	}
}

void Ball::ApplyPosbased_vel(Time prevtime,Time latestime,float conf){
	bool valid;
	Vector evel_posbased;
	evel_posbased = posbased_vel_estimate(prevtime, latestime, valid);
	if(valid){
		gvel.Setdata(evel_posbased, latestime);
		vel_time = latestime;
		speed_conf = conf;
		DoLog(LOG_UPDATE,"t%d, Posbased vel applied", latestime);
	}else{
		DoLog(LOG_UPDATE,"Why posbased_vel Invalid (%.2f %.2f)", evel_posbased.x, evel_posbased.y);
	}
	velgap.Setdata(posgap.ChkData(prevtime) + posgap.ChkData(latestime), latestime);
	velgap_time = latestime;
	DoLog(LOG_UPDATE, "reset velgap (%.2f, %.2f)",velgap.Data(latestime).x, velgap.Data(latestime).y);
}

//decide whether the hear info is useful
bool Ball::Isupdate_heard(){
	//no hear info
	if(!heard_pos) return false;
	
	if(IsValidObject()){
		if(pos.dist(heardpos_info.data.mean) > heardpos_info.data.delta + pos_delta + 2.7f){
			DoLog(LOG_HEAR, "Hearbug %d known pos(%.1f, %.1f), delta%.1f while hearing pos(%.1f, %.1f), delta%.1f",
				heardpos_info.time, pos.x, pos.y, pos_delta,
				heardpos_info.data.mean.x, heardpos_info.data.mean.y, heardpos_info.data.delta);
		}
	}
	DoLog(LOG_HEAR, "ball heard pos(%.1f, %.1f, delta%.1f), delay%d old_delta%.1f", 
		heardpos_info.data.mean.x, heardpos_info.data.mean.y, heardpos_info.data.delta,
		situation.CurrentTime - heardpos_info.time, pos_deltas.Data(pos_delta_time));
	if(heardpass.time == heardpos_info.time){
		DoLog(LOG_HEAR, "heard passing");
	} 
	//old info is not valid
	if(!pos_valid() || !IsValidObject() || Isforgot()){		
		find_kick = true;
		return true;
	}
	//I have newer info already
	if(Last_seen_time() > heardpos_info.time){
		if(heardpass.time != heardpos_info.time || seenchinfo.IsDataKnown(Last_seen_time())){
			return false;
		}else{
			//but ball's vel isn't be seen for too far.			
			find_kick = true;
			return true;
		}
	}
	
	//hear pass info
	if(heardpass.time == heardpos_info.time && heardpos_info.time >= Last_seen_time()){
		find_kick = true;
		return true;
	}
	//old decision by criver
	if (pos_conf > 0.9f || distance < 2.0f){
		//it is critical for me, if the confidence of ball is greater than 0.8f and it ball is within 2 meter of my reach
		if (heardvel_info.time != heardpos_info.time)	return false;
		if (ball.global_vel.dist2(heardvel_info.data.mean) < Sqr(0.6f)) return false;
	}


	//my info is clear enough
	if(IsLastSeen() && !OtherKick(update_seen_time)) return false;

	if(!IsLastSeen() && (heardvel_info.time - heardvel_info.data.delta)  > update_seen_time && IscplayerKnown(update_seen_time)){
		//ܸձ߹
		Vector rel_ball_pos = ball.Pos(update_seen_time) - Self.Pos(update_seen_time);
		if(closestrpos(update_seen_time).mod() - rel_ball_pos.mod() * ServerParam::dist_qstep * 0.6f < ServerParam::kickable_area){
			if(!find_kick){
				estimate(heardvel_info.time);
				find_kick = heardvel_info.data.mean.dist2(ball.Vel(heardvel_info.time)) > Sqr(0.6f);
			}
			return true;
		}
	}
	//convert hear delta to pos gap in my system
	float relateive_delta = heardpos_info.data.delta + CP_coordinate_gap_between_agents;

	estimategap(heardpos_info.time);
	if(Sqr(relateive_delta) < posgap.Data(heardpos_info.time).mod2()){
		//rude deal
		find_kick = true;
		return true;
	}else{
		return false;
	}
	//һ㻹ûж, λʱԼ֪, Ȼdelta>posgap, 
	//м߹,Ӧ.
}

void Ball::update_heard(Time time){
	MobileObject::update_heard(time); 

	if(isupdateheard){
		//take care, here we set player's iscontrolball		
		int No;
		for(No = 1; No <= SP_team_size; No++){
			MyPlayer(No).Set_Iscontrolball_Fromheard(heardpos_info.time);
			TheirPlayer(No).Set_Iscontrolball_Fromheard(heardpos_info.time);
			UnknownTheirPlayerList.Set_IsControlball_FromHeard(heardpos_info.time);
			UnknownMyPlayerList.Set_IsControlball_FromHeard(heardpos_info.time);
			TeamlessPlayerList.Set_IsControlball_FromHeard(heardpos_info.time);
		}
		updated_holder = true;
		if (heard_vel && heardvel_info.data.delta <= VelDelay() ){
			//λþȷȸ,ΪٶȵҲ
			gvel.Setdata(heardvel_info.data.mean, heardvel_info.time);
			vel_time = heardvel_info.time;
			speed_conf = (float)pow(conf_decay, int(heardvel_info.data.delta));

			original_veltime = heardvel_info.time - int(heardvel_info.data.delta);
			newvel = true;
			DoLog(LOG_HEAR, "ball vel %d, (%.1f,%.1f)", heardvel_info.time, 
				heardvel_info.data.mean.x, heardvel_info.data.mean.y);
		}
	}
}

void Ball::Prior_update(){
	find_kick = false;
	updated_holder = false;
}

void Ball::update_end(){
	MobileObject::update_end();

	if(Last_seen_time() < Self.Last_seen_time() && distance < ServerParam::feel_distance){
		if(ball.Pos(Self.Last_seen_time()).dist(Self.Pos(Self.Last_seen_time())) + ball.posgap.ChkData(Self.Last_seen_time()).mod() < ServerParam::feel_distance){
			forget();
		}
	}
	pos_delta = posgap.ChkData(situation.CurrentTime).mod();
	if(distance > ClientParam::kickable_area || !pos_valid() || !IsValidObject()){
		kick_able = false;
	}else if(Last_seen_time() >= Self.Last_seen_time()){
		kick_able = true;
	}else if(pos_delta > 2.0f){
		kick_able = false;
	}else if(ball.Pos(Self.Last_seen_time()).dist(Self.Pos(Self.Last_seen_time())) + 
		ball.posgap.ChkData(Self.Last_seen_time()).mod() > ServerParam::feel_distance){
		kick_able = true;	
	}else{
		kick_able = false;
	}
	if(distance < ClientParam::kickable_area && !kick_able){
		if(!pos_valid()) DoLog(LOG_UPDATE, "kick disable for pos invalid");
		else if(!IsValidObject()) DoLog(LOG_UPDATE, "kick disable for Rs invalid");
		else{
			DoLog(LOG_UPDATE, "kick disable for ball( %d) is unseen at %d", Last_seen_time(), Self.Last_seen_time());
		}
	}
	theirgoalangle = (PitchInfo.theirgoal - pos).Angle();
	mygoalangle = (PitchInfo.mygoal - pos).Angle();
	DoLog(LOG_UPDATE, "ball pos(%.2f, %.2f), vel(%.2f, %.2f)", pos.x, pos.y, global_vel.x, global_vel.y);
}
 
void Ball::Collisionvel(Time time,float conf){
	Vector vel = Skill::action.SummarizeKickeffectVel(Vel(vel_time),vel_time,time,speed_decay,max_speed);
	gvel.Setdata(vel * -0.1f, time);
	vel_time = time;
	speed_conf = conf;
	DoLog(LOG_UPDATE,"coll handle t%.0f,v(%.2f,%.2f)",(float)time,Vel(time).x,Vel(time).y);
}

bool Ball::OtherKick(Time time){
	if(IscrposKnown(time)){
		if(closestrpos(time).mod() < ServerParam::kickable_area + 0.05f )
			return true;
	}
	return false;
}

 bool Ball::OtherKick(Time prevtime,Time lastime){
	for(Time t=prevtime;t<=lastime;t++){
		if(OtherKick(t)) return true;
	}
	return false;
}

bool Ball::IsposError(Vector allowposgap, Vector predictgap,float& gapmod){
	if(fabs(predictgap.x) > allowposgap.x || fabs(predictgap.y) > allowposgap.y){
		gapmod = fabs(predictgap.x) > allowposgap.x ? Sqr((float)fabs(predictgap.x)-allowposgap.x) : 0;
		gapmod += fabs(predictgap.y) > allowposgap.y ? Sqr((float)fabs(predictgap.y)-allowposgap.y) : 0;
		gapmod = (float)sqrt(gapmod / allowposgap.mod());
		return true;
	}else{
		gapmod = 0;
		return false;
	}
}

Vector Ball::posbased_vel_estimate(Time prev_time, Time last_time, bool& valid){
	/* estimate by postion difference */
	valid = false;
	Vector gap = Pos(last_time) - Pos(prev_time);
	Vector evel;
	gap -= Self.SummarizePosGap(prev_time, last_time);

	DoLog(LOG_UPDATE,"prev pos (%.2f %.2f) last pos (%.2f %.2f) gap (%.2f %.2f)", Pos(prev_time).x, Pos(prev_time).y,
		Pos(last_time).x, Pos(last_time).y, gap.x, gap.y);
	//if (gap.mod() < (float)fabs(last_time - prev_time) * max_speed){
	Vector kickeffectpos = Skill::action.SummarizeKickeffectPos(prev_time, last_time, speed_decay);
	evel = (gap - kickeffectpos) * (1 - speed_decay) / (1 - Exp(speed_decay, (float)(last_time - prev_time)));
	//evel is now that at time prev_time
	if(evel.mod() > max_speed){
		evel = evel / evel.mod() * max_speed;
	}
	evel = Skill::action.SummarizeKickeffectVel(evel, prev_time, last_time, speed_decay, max_speed);
	valid = true;
	return evel;
}

void Ball::estimate(Time time){
	MobileObject::estimate(time);
	estimategap(time);
}

void Ball::estimate_pos(Time time){
	MobileObject::estimate_pos(time);
}

void Ball::CollisionHandle(Time time){
	DoLog(LOG_UPDATE,"Coolisionhandle");
	estimate_pos(time);
	if (speed_valid()){
		Vector rpos = Pos(time) - Self.Pos(time);
		float r = ServerParam::ball_size + ServerParam::player_size;
		float d = rpos.mod();
		if (d>=r) return; 
		AngleDeg th = float(fabs(NormalizeAngle(rpos.Angle() - Self.Bodyfacing(time -1) - 180.0f)));
		float l1 = d * Cos(th);
		float h = d * Sin(th);
		float cosp = h / r;
		float sinp = (float)sqrt(1.0 - Sqr(cosp));
		float l2 = r * sinp;
		gpos.Data(time) += Vel(time) *float( speed_decay * (-(l1+l2) / Max(Vel(time).mod() * speed_decay, 0.000001f)));
		gvel.Data(time) *= -0.1f;
		vel_time = time;
		speed_conf = min_valid_conf / Sqr(conf_decay);
	}
}

bool Ball::catchable(){
	if(Self.Is_goalie)
		return bool(pos_conf > 0.8f && distance < ServerParam::catch_area_l && PitchInfo.my_penaltyarea.IsWithin(Self.pos) && Skill::action.cancatch(situation.CurrentTime));
	else
		return bool(pos_conf > 0.8f && distance < ServerParam::catch_area_l && PitchInfo.my_penaltyarea.IsWithin(pos) && Skill::action.cancatch(situation.CurrentTime));
}

bool Ball::shootable(){
	return bool(FieldInfo.shootablearea.IsWithin(pos) && kickable());
}

bool Ball::kickable(const Vector& selfpos){
	return bool(pos_valid() && ball.pos.dist2(selfpos) < Sqr(Self.kickable_area ));
}

bool Ball::catchable(const Vector& selfpos){
	return bool(pos_valid() && pos.dist2(selfpos) < Sqr(ServerParam::catch_area_l) && PitchInfo.my_penaltyarea.IsWithin(selfpos) && Skill::action.cancatch(situation.CurrentTime));
}

bool Ball::shootable(const Vector& ballpos){
	return FieldInfo.shootablearea.IsWithin(ballpos);
}

void Ball::Setball_CPlayer(UNum idx,Vector pos,Time time){
	if(idx > 0 && idx <= 2*SP_team_size)
		ClosestPlayer.Setdata(idx, time);
	else
		ClosestPlayer.Setdata(idx, time-1);
	ClosestRPos.Setdata(pos-Pos(time),time);
}

bool Ball::IsOtherKick(){
	if(IsOtherKick(situation.CurrentTime) || IsOtherKick(situation.CurrentTime - 1)){
		return true;
	}else
		return false;
}

bool Ball::IsOtherKick(Time t){
	if(error_idx.IsDataKnown(t) && error_idx.Data(t) > 0.01f )
		return true;
	else
		return false;
}
