#include "player.h"
#include "worldmodel.h"
#include "perceptron.h"
#include "skill.h"
#include "log.h"
#include "agent.h"


PlayerType::PlayerType(){
	SetDefaultType();
}

void PlayerType::SetDefaultType(){
	player_speed_max = ServerParam::player_speed_max;	
	stamina_inc_max = ServerParam::stamina_inc;
	player_decay = ServerParam::player_decay;
	inertia_moment = ServerParam::inertia_moment;
	dash_power_rate = ServerParam::dash_power_rate;
	player_size = ServerParam::player_size;
	kickable_margin= ServerParam::kickable_margin;
	kick_rand = ServerParam::player_rand;
	extra_stamina = 0;
	effort_max = 1.0f;
	effort_min = ServerParam::effort_min;
	
	kickable_area = player_size + kickable_margin + ServerParam::ball_size;
	player_speed_max = Min(player_speed_max, ServerParam::max_power*dash_power_rate / ( 1-player_decay)); 
}

void PlayerType::CopyType(int i){
	if(i<0 || i>=SP_player_types) return;
	
	player_speed_max = playertypes[i].player_speed_max;
	stamina_inc_max = playertypes[i].stamina_inc_max;
	player_decay = playertypes[i].player_decay;
	inertia_moment = playertypes[i].inertia_moment;
	dash_power_rate = playertypes[i].dash_power_rate;
	player_size = playertypes[i].player_size;
	kickable_margin = playertypes[i].kickable_margin;
	kick_rand = playertypes[i].kick_rand;
	extra_stamina = playertypes[i].extra_stamina;
	effort_max = playertypes[i].effort_max;
	effort_min = playertypes[i].effort_min;

	kickable_area = player_size + kickable_margin + ServerParam::ball_size;
	player_speed_max = Min(player_speed_max, ServerParam::max_power*dash_power_rate / ( 1-player_decay)); 
}

/***************    Role   ******************/
Role::Role(float attackness, float leftness, float aggressiveness, float creative){
	SetRole(attackness, leftness, aggressiveness, creative);
}

void Role::SetRole(float attackness, float leftness, float aggressiveness, float creative){
	this->attackness = attackness;
	this->leftness = leftness;
	this->aggressiveness = aggressiveness;
	this->creativeness = creativeness;
}

bool Role::RoleMatched(Role& role){
	return bool(attackness == role.attackness && leftness == role.leftness);
}
/**************************   Player    ******************************/
Player::Player(){
	MobileObject::MobileObject();
	InsideNO = -1;
	headfacing = 0;
	bodyfacing = 0;
	IT_inf.IT_cycles = InfCycles;
	Is_goalie = false;
	side = Side_Unknown;
	Iscontrolball = false;
	original_postime = 0;
	original_veltime = 0;

	//only useful for dynamic constructed player
	set_conf_param(ClientParam::conf_max, ClientParam::conf_decay, ClientParam::min_valid_conf);
	set_objtype(OBJ_Player);
	ChangeType(0);
}

void Player::Reset(Time time){
	vel_time = time - 1;
	gvel.Setdata(0, time -1);
	speed_conf = 0;
	update_seen_time = Max(time - 1, 0);

	pos_time = time -1;
	gpos.Setdata(0, pos_time);

	seenposinfo.TimedData(time - 1).time = -1;
	seenchinfo.TimedData(time -1).time = -1;
}

// to decide whether consistent with unknown player in sight info.
// original pos and delta is taken as reliable knowledge
bool Player::IsSamePlayer(VisualUnkPlayer& newp){
	float dist, delta;
	int timegap;

	if(gpos.IsDataKnown(original_postime)){ 
		dist = Pos(original_postime).dist(newp.pos);
		delta = pos_deltas.Data(original_postime);
		timegap = abs(newp.t - original_postime);
	}else{
		dist = pos.dist(newp.pos);
		delta = dist * 0.5f * ServerParam::dist_qstep;
		timegap = 8;
	}

	if(dist > ClientParam::too_different_distance) return false;
	float allowdis = newp.seen_distance() * ServerParam::dist_qstep + timegap * max_speed + delta + Self.SummarizePosGap(original_postime, newp.t).mod();
	if(Self.SelfColl(original_postime, newp.t)){
		allowdis += 1.0f;
	}
	if(dist < allowdis)
		return true;
	else
		return false;
}


void Player::estimate_domain(Time time){
	if(domain_time >= time) return;
	if(!domain.IsDataKnown(domain_time) || !domain.Data(domain_time).IsValid()) return;

	if(sensory.NewSight && !seen && domain_time == sensory.LastSightTime){
#ifdef _WINDOWS
		domain.Data(domain_time).MinusAngle(NoiseData(Self.Headfacing(domain_time),
					0.5f*Self.MyViewWidth(Self.viewwidth.Data(domain_time))));
#else // _LINUX
		NoiseData tmpnoise = NoiseData(Self.Headfacing(domain_time),
                                        0.5f*Self.MyViewWidth(Self.viewwidth.Data(domain_time)));
		domain.Data(domain_time).MinusAngle(tmpnoise);
#endif
		domain.Data(domain_time).MinusRadius(ServerParam::feel_distance);
	}
	while(domain_time < time){
		domain_time ++;
		domain.Data(domain_time) = domain.Data(domain_time-1);
		domain.TimedData(domain_time).time = domain_time;
		domain.Data(domain_time).Diffuse(1,max_speed);
		domain.Data(domain_time).ShiftCenter(Self.Pos(domain_time) - Self.Pos(domain_time-1));
		if(sensory.rec_sight_info.ChkData(domain_time)){
			if(seenposinfo.IsDataKnown(domain_time)){
			//ע,б, ͲӦýģʽ, 
				DoLog("Error estimate Domain");				
				domain.Data(domain_time).Set(seen_distance(domain_time), 
					0.5f*seen_distance(domain_time)*ServerParam::dist_qstep,
					seen_rel_angle(domain_time)+Self.Headfacing(domain_time), 
					Rad2Deg(0.5f*ServerParam::dir_qstep));
			}else{	
			//Լû
#ifdef _WINDOWS
				domain.Data(domain_time).MinusAngle(NoiseData(Self.Headfacing(domain_time),
					0.5f*Self.MyViewWidth(Self.viewwidth.Data(domain_time))));
#else // _LINUX
				NoiseData  tmpnoise;
				tmpnoise = NoiseData(Self.Headfacing(domain_time),
                                        0.5f*Self.MyViewWidth(Self.viewwidth.Data(domain_time)));
				domain.Data(domain_time).MinusAngle(tmpnoise);
#endif
				domain.Data(domain_time).MinusRadius(ServerParam::feel_distance);
			}
		}
	}
}

void Player::update(Time time){
	if(InsideNO == Agent::MyNumber) return;

	MobileObject::update(time);
#ifdef	_Up_FState
	Iscontrolball = Pos(time).dist2(ball.Pos(time)) < Sqr(kickable_area);
#else
	if(ball.updated_holder) return;
	//check if it control the ball
	if (ball.IsSeen() && IsValidObject()){
	//see ball
		Iscontrolball = (Pos(ball.seen_time()).dist2(ball.Pos(ball.seen_time())) < Sqr(kickable_area));
	}else if(!ball.IsSeen() && IsSeen()){//(IsSeen(time)){
		//Iscontrolball = false;	//ֻ,û
		if(Iscontrolball){
			int t = seen_time();
			float edge_gap = Self.MyViewWidth(Self.viewwidth.Data(t))/2 - (float)fabs(seen_rel_angle());
			float allow_diff = ASin(kickable_area / (seen_distance()+ 0.5f));
			if(edge_gap > 2.9f || (allow_diff - edge_gap) < 0.5f 
				|| (seen_distance() <= ServerParam::feel_distance && !seenchinfo.IsDataKnown(t))){
				Iscontrolball = false;
				DoLog("controlball fix, %d (%.1f, %.1f, %.1f)", InsideNO, edge_gap, allow_diff, edge_gap - allow_diff);
			}else{
				DoLog("controlball bug see player %d withou ball(%.1f, %.1f, %.1f), how can he control ball?", InsideNO, edge_gap, allow_diff, edge_gap - allow_diff);
			}
		}
	}
#endif
}

void Player::update_seen(Time time){
	MobileObject::update_seen(time);
	/*******domain - ԱĻΧ**********/
	int t = update_seen_time;//
	if(seen){
		domain.Data(t).Set(seen_distance(t), 0.5f*seen_distance(t)*ServerParam::dist_qstep,
			seen_rel_angle(t)+Self.Headfacing(t), Rad2Deg(0.5f*ServerParam::dir_qstep));
		domain.TimedData(t).time = t;
		domain_time = t;
	}
}

void Player::update_heard(Time time){
	MobileObject::update_heard(time);

	if(isupdateheard){
		//Update a player's domain according to heard info which is a circle.
		estimate_domain(heardpos_info.time);
		Vector relpos = heardpos_info.data.mean - Self.Pos(heardpos_info.time);

		NoiseData radius(relpos.mod(), heardpos_info.data.delta);
		NoiseData angle;
		angle.mean = relpos.Angle();
		if(heardpos_info.data.delta >= radius.mean){
			//when the circle covers myself
			angle.delta = 180.0f;
		}else{
			angle.delta = ASin(heardpos_info.data.delta / radius.mean);
		}
		if(domain.IsDataKnown(heardpos_info.time) && domain.Data(heardpos_info.time).IsValid() ){
			domain.Data(heardpos_info.time).Intersect(radius, angle);
		}else{
			domain.Data(heardpos_info.time).Set(radius.mean, radius.delta,
				angle.mean, angle.delta);
			domain.TimedData(heardpos_info.time).time = heardpos_info.time;
		}
		domain_time = heardpos_info.time;
		DoLog(LOG_HEAR, "heard%d at %d(%.1f,%.1f) delta %.1f", InsideNO, heardpos_info.time, heardpos_info.data.mean.x, 
			heardpos_info.data.mean.y, heardpos_info.data.delta);
	}
}

void Player::update_end(){
	//update bodyfacing, headfc
	if(InsideNO != Agent::MyNumber){
#ifdef _Up_FState
		if(sensory.HasFullState()){
			if(acc_bodyangle.IsDataKnown(situation.CurrentTime)){
				bodyfc.Setdata(acc_bodyangle.Data(situation.CurrentTime), situation.CurrentTime);
				if(acc_neckangle.IsDataKnown(situation.CurrentTime)){
					headfc.Setdata(NormalizeAngle(acc_neckangle.Data(situation.CurrentTime)
						+ bodyfc.Data(situation.CurrentTime)), situation.CurrentTime);
				}
				bodyfc_guessed = false;
			}
		}
#else
		//
		if(newpos){
			bodyfc_guessed = false;
			if(fcinfo.time == situation.CurrentTime){
				bodyfc.Setdata(NormalizeAngle(Self.headfacing+fcinfo.data.x), situation.CurrentTime);
				headfc.Setdata(NormalizeAngle(Self.headfacing+fcinfo.data.y), situation.CurrentTime);
			}else if(IsSeenMoving(situation.CurrentTime)){
				//۲쵽Ա˶ٶ<0.1fʱ, Ϊ峯ж	
				bodyfc.Setdata(Vel(situation.CurrentTime).Angle(), situation.CurrentTime); 
				if(Vel(situation.CurrentTime).mod() < 0.1f){
					bodyfc_guessed = true;			
				}
			}else if(seenposinfo.IsDataKnown(situation.CurrentTime - 1) && distance < 5.0){
				//λòλ
				Vector vel = Pos(situation.CurrentTime) - Pos(situation.CurrentTime -1);
				vel -= Self.SummarizePosGap(situation.CurrentTime-1, situation.CurrentTime);				
				bodyfc.Setdata(vel.Angle(), situation.CurrentTime);
				
				if (vel.mod() > max_speed){
					vel = vel / vel.mod() * max_speed;
				}
				if(vel.mod() <  (distance * 0.05f + 0.15f)){
					//λùƳ˶ٶ<0.2fʱ, Ϊ峯ж	
					bodyfc_guessed = true;
				}
				vel *= speed_decay;
				gvel.Setdata(vel, situation.CurrentTime);
				vel_time = situation.CurrentTime;
				original_veltime = vel_time;
				speed_conf = max_conf;				
			}else{
				bodyfc_guessed = true;
				if(gpos.IsDataKnown(situation.CurrentTime - 1)){
					Vector vel = Pos(situation.CurrentTime) - Pos(situation.CurrentTime -1);					
					bodyfc.Setdata(vel.Angle(), situation.CurrentTime);
				}
			}
		}else{
			bodyfc_guessed = true;
		}
		estimate_domain(situation.CurrentTime);
		pos_delta_valid = domain.IsDataKnown(situation.CurrentTime) && domain.Data(situation.CurrentTime).IsValid();
		estimate_pos_delta(situation.CurrentTime);
#endif
	}else{
		pos_delta_valid = true;
		pos_deltas.Setdata(0, situation.CurrentTime);  
	}

	MobileObject::update_end();

	rel_pos_2_ball = pos - ball.pos;
	balldist = rel_pos_2_ball.mod();
	ballangle = (ball.pos - pos).Angle();

	if(!IsValidObject()) return;

	Vector domain_pos = domain.Data(situation.CurrentTime).GetCenterPos(); 
	DoLog(LOG_UPDATE, "No%d, pos(%.1f, %.1f), domain_pos(%.1f, %.1f), delta%.1f, conf%.2f",
		InsideNO, pos.x, pos.y, domain_pos.x, domain_pos.y, pos_delta, pos_conf);
	if(bodyfc.IsDataKnown(situation.CurrentTime)){
		bodyfacing = bodyfc.Data(situation.CurrentTime);
		if(!bodyfc_guessed){
			DoLog(LOG_UPDATE, "NO %d, bodyfacing%.2f, speed%.2f, Angle%.2f", InsideNO, bodyfacing,
				speed, vel_angle);
		}else{
			DoLog(LOG_UPDATE, "NO %d, bodyfacing%.2f(guess), speed%.2f, Angle%.2f", InsideNO, bodyfacing,
				speed, vel_angle);
		}
	}
}

//decide whether hear info is useful for me.
bool Player::Isupdate_heard(){	
	//no hear info
	if(!heard_pos) return false;
	
	if(IsValidObject()){
		if(pos.dist(heardpos_info.data.mean) > heardpos_info.data.delta + pos_delta + 1.0f){
			DoLog(LOG_HEAR, "Hearbug %d known pos(%.1f, %.1f), delta%.1f while hearing pos(%.1f, %.1f), delta%.1f",
				heardpos_info.time, pos.x, pos.y, pos_delta,
				heardpos_info.data.mean.x, heardpos_info.data.mean.y, heardpos_info.data.delta);
		}
	}
	DoLog(LOG_HEAR, "%d (%.1f, %.1f, delta%.1f), delay%d, old_delta%.1f",InsideNO, 
		heardpos_info.data.mean.x, heardpos_info.data.mean.y, heardpos_info.data.delta,
		situation.CurrentTime - heardpos_info.time, pos_deltas.Data(pos_delta_time));
	//old info is not valid
	if(!pos_valid() || !IsValidObject() || Isforgot()) return true;

	float guess_delta;
	estimate_pos_delta(situation.CurrentTime);
	guess_delta = heardpos_info.data.delta + (situation.CurrentTime 
		- heardpos_info.time) * max_speed + CP_coordinate_gap_between_agents;


	if(guess_delta < pos_deltas.Data(pos_delta_time)){
		return true;
	}else{
		return false;
	}
}

AngleDeg Player::Bodyfacing(Time time){
	return bodyfc.IsDataKnown(time) ? bodyfc.Data(time) : bodyfacing;
}

AngleDeg Player::Headfacing(Time time){
	return headfc.IsDataKnown(time) ? headfc.Data(time) : headfacing;
}

AngleDeg Player::Headangle(Time time){
	return headang.IsDataKnown(time) ? headang.Data(time) : head_angle;
}

void Player::set_fcinfo(float facedir, float neckdir, Time time){
	fcinfo = TimedData<Vector>(Vector(facedir, neckdir), time);
}

void Player::set_arm_tackle(float armdir, bool Istackle, Time time){
	seenarm.Setdata(armdir, time);
	seentackle.Setdata(Istackle, time);
}

void Player::set_inside_no(UNum No){
	this->InsideNO = No;
}

void Player::set_side(bool Is_myside){
	if (Is_myside){
		side = Side_My;
	}
	else{
		side = Side_Opp;
	}
}

void Player::set_side(PlayerSide side){
	this->side = side;
}

void Player::ChangeType(int type){
	CopyType(type);
	max_speed = player_speed_max;
	speed_decay = player_decay;
}

void Player::Set_Iscontrolball_Fromheard(Time time){
	if(heard_iscontrolball.time == time){
		Iscontrolball = heard_iscontrolball.data; 
	}
}

bool Player::IsStrained(){
	return bool(stamina < 0.5f * ServerParam::stamina_max);
}

bool Player::IsExtremStrained(){
	return bool(stamina < ServerParam::stamina_max * ServerParam::recover_dec_thr);
}

bool Player::IsUniformNo(UNum No){
	return No >0 && No <= 2*SP_team_size;
}

bool Player::IsValidNo(UNum No){
	return No > 0 && No <= 7 * SP_team_size;
}

bool Player::IsMyPlayer(UNum No){
	if (No <= 0) return false;
	int seg = (No -1 ) / SP_team_size;
	int rem = (No -1)  % SP_team_size;
	switch(seg){
	case 0:
		return true;
	case 2:
	case 3:
		return UnknownMyPlayerList.IsIdxValid(rem);
//	case 6:
//		return TeamlessPlayerList.IsIdxValid(rem);
		
	}
	return false;
}

bool Player::IsTheirPlayer(UNum No){
	if (No <= 0) return false;
	int seg = (No -1 ) / SP_team_size;
	int rem = (No -1)  % SP_team_size;
	switch(seg){
	case 0:
		return false;
	case 1:
		return true;
	case 4:
	case 5:
		return UnknownTheirPlayerList.IsIdxValid(rem);
	case 6:
		return TeamlessPlayerList.IsIdxValid(rem);		
	}
	return false;
}

bool Player::IsUnknownPlayer(UNum No){
	return bool(No <= 0 || No > 2 * SP_team_size);
}



/**************************   Self   *********************************/
MySelf::MySelf(){
	pos_delta = 0;
	pos_conf = 1.0f;
	bodyfc_guessed = false;
	pos_delta_valid = true;
	updated_sight_time = -1; 
}

void MySelf::update(Time time){
/*********	time is assumed to be CurrentTime here	********/
	if (sensory.SensedInfKnown(time)){
		update_from_senseinfo(sensory.GetSenseBodyInfo(time));
	}else{
		estimate_headangle(time);
		estimate_stamina(time);
		estimate_effort(time);
	}

	update_seen_facing(time);
	estimate_facing(time);

	update_vel(time);// vel should be updated before pos
	update_seen_pos(time);
	estimate_pos(time);

	pos = Pos(time);
	global_vel = Vel(time);
	speed = global_vel.mod();
	vel_angle = global_vel.Angle();
	bodyfacing = Bodyfacing(time);
	headfacing = Headfacing(time);
	head_angle = Headangle(time);
	stamina = Stamina(time);
	effort = Effort(time);
	pos_time = time;
	vel_time = time;
	Focus = FocusPlayers.ChkData(time);
}

void MySelf::update_from_senseinfo(SenseBodyInfo & sb){
	viewwidth.Setdata(sb.view_width, sb.time);
	viewqual.Setdata(sb.view_qual, sb.time);
	stmn.Setdata(sb.stamina, sb.time);
	spd.Setdata(sb.speed < max_speed * speed_decay ? sb.speed : max_speed * speed_decay, sb.time);
	spdang.Setdata(sb.rel_spd_angle, sb.time);
	headang.Setdata(sb.head_angle, sb.time);
	efft.Setdata(sb.effort, sb.time);
	FocusPlayers.Setdata(sb.Focus, sb.time);
}

void MySelf::estimate_headangle(Time time){
	if (Skill::action.Valid(CMD_turn_neck)){
		headang.Setdata(Headangle(situation.LastCycle()) + Skill::action.TurnNeckEffect(situation.LastCycle()), time);
	}
}

void MySelf::estimate_stamina(Time time){
	stmn.Setdata(Stamina(situation.LastCycle()) - Skill::action.UsedPower(situation.LastCycle()), time);
}

void MySelf::estimate_effort(Time time){
	efft.Setdata(Effort(situation.LastCycle()), time);
}

void MySelf::update_seen_facing(Time time){
#ifdef	_Up_FState
	if(sensory.HasFullState() && acc_bodyangle.IsDataKnown(time)){
		bodyfc.Setdata(NormalizeAngle(acc_bodyangle.Data(time)), time);
		headfc.Setdata(NormalizeAngle(acc_bodyangle.Data(time)+acc_neckangle.Data(time)), time); 
		return;
	} 
#endif
	if (!sensory.NewSight) return;

/******	update facings from seen	************/
	int last_update_seen_time = Max(update_seen_time, time -2);
	int t;
	for(t = last_update_seen_time + 1; t <= sensory.LastSightTime; t ++){
		if (sensory.num_seenlines.ChkData(t) > 0){
			int line = sensory.SeenLines.Data(0, t);
			float rel_angle_2_line = PitchInfo.Lines[line].seen_rel_angle(t);
			if (rel_angle_2_line < 0){
				rel_angle_2_line += 180.0f;
			}

			headfc.Setdata(NormalizeAngle(PitchInfo.Lines[line].global_angle - rel_angle_2_line), t);
			bodyfc.Setdata(NormalizeAngle(Headfacing(t) - Headangle(t)), t);
		}
		else if (sensory.num_seenmarkers.ChkData(t) >= 2 ){
			//find two markers with known distance
			MarkerType marker1, marker2;
			marker1 = sensory.SeenMarkers.Data(0, t);
			marker2 = sensory.SeenMarkers.Data(1, t);

			PitchInfo.Markers[marker1].rel_pos = Polar2Vector(PitchInfo.Markers[marker1].seen_distance(t), PitchInfo.Markers[marker1].seen_rel_angle(t));
			PitchInfo.Markers[marker2].rel_pos = Polar2Vector(PitchInfo.Markers[marker2].seen_distance(t), PitchInfo.Markers[marker2].seen_rel_angle(t));
			Vector gap1 = PitchInfo.Markers[marker1].rel_pos - PitchInfo.Markers[marker2].rel_pos;
			Vector gap2 = PitchInfo.Markers[marker1].pos - PitchInfo.Markers[marker2].pos;
			headfc.Setdata(NormalizeAngle(gap2.Angle() - gap1.Angle()),	t);
			bodyfc.Setdata(NormalizeAngle(Headfacing(t) - Headangle(t)), t);
		}
	}
}

void MySelf::estimate_facing(Time time){
	if(!IsBodyfacingKnown(time)){
		if (!IsBodyfacingKnown(situation.LastCycle())){
			/********	must miss some cycles	************/
			bodyfc.Setdata(NormalizeAngle(Bodyfacing(pos_time) + Skill::action.SummarizeTurneffect(pos_time, situation.LastCycle()))
				, situation.LastCycle());
			if(IsHeadAngleKnown(situation.LastCycle())){
				headang.Setdata(NormalizeAngle(Headangle(pos_time) + Skill::action.SummarizeTurnneckeffect(pos_time, situation.LastCycle()))
					, situation.LastCycle());
			}
			headfc.Setdata(NormalizeAngle(Bodyfacing(situation.LastCycle()) + Headangle(situation.LastCycle())), 
					situation.LastCycle());
		}

		bodyfc.Setdata(NormalizeAngle(Bodyfacing(situation.LastCycle()) + Skill::action.TurnEffect(situation.LastCycle())), time);
		headfc.Setdata(NormalizeAngle(Bodyfacing(time) + Headangle(time)), time); 
	}
}

void MySelf::update_vel(Time time){
#ifdef	_Up_FState
	if(sensory.HasFullState() && acc_vel.IsDataKnown(time)){
		gvel.Setdata(acc_vel.Data(time), time);
		vel_time = time;
		return;
	} 
#endif

	Vector evel = Skill::action.SummarizeDasheffectVel(Vel(vel_time), vel_time, time, speed_decay, max_speed);

	if (sensory.SensedInfKnown(time)){
		gvel.Setdata(Polar2Vector(Speed(time), SpeedAngle(time) + Headfacing(time)), time);

		float Velgap =(float)fabs(gvel.Data(time).mod() - evel.mod());

		Vector svgap = gvel.Data(time) - evel;

		if(Velgap >ClientParam::selfvel_mingap && Speed(time) <= max_speed * 0.11f * speed_decay){
			DoLog(LOG_COLLIDE,"self coll Velgap%.2f,Vector Gap(%.2f,%.2f)",Velgap,svgap.x,svgap.y);
			selfcoll.Setdata(true,time);
		}else
			selfcoll.Setdata(false,time);
	}else{
		/*estimate vel*/
		gvel.Setdata(evel, time);
	}
	vel_time = time;
}

void MySelf::update_seen_pos(Time time){
#ifdef	_Up_FState
	if(sensory.HasFullState() && acc_vel.IsDataKnown(time)){
		gpos.Setdata(acc_pos.Data(time), time);
		pos_time = time;
		newpos = true;
		update_seen_time = time;
		pos_gap.Setdata(0, time);
		return;
	}
#endif
/********	Update Pos from seen	**************/
	if(!sensory.NewSight) return;

	int last_update_seen_time = Max(update_seen_time, time -2);
	MarkerType closestmarker;
	int t;
	for(t = last_update_seen_time + 1; t <= sensory.LastSightTime; t ++){
		if (sensory.num_seenmarkers.ChkData(t) > 0){
			/***********	calculate pos gap	***************/
 			closestmarker = sensory.GetClosestMarker(t);
			if (closestmarker != No_Marker){
				PitchInfo.Markers[closestmarker].rel_pos = Polar2Vector(PitchInfo.Markers[closestmarker].seen_distance(t),
					Headfacing(t) + PitchInfo.Markers[closestmarker].seen_rel_angle(t));
				gpos.Setdata(PitchInfo.Markers[closestmarker].pos - PitchInfo.Markers[closestmarker].rel_pos, t);
				DoLog(LOG_UPDATE, "Player Pos Updated (%.2f %.2f) at %d", Pos(t).x, Pos(t).y, t);

				if(pos_time < t -1){
					estimate_pos(t-1);
				}
				Vector posbk = Pos(t-1) + Vel(t) / speed_decay;
				pos_gap.Setdata(Pos(t) - posbk, t);
				if(pos_gap.Data(t).mod() > 4.0f){
					DoLog("I'm moved");
					selfmove.Setdata(true, t);
					pos_gap.Setdata(0, t);
				}else if(selfcoll.ChkData(t)){
					pos_gap.Setdata(0, t);
				} 
				DoLog(LOG_UPDATE, "PrevCoord (%.2f %.2f)", posbk.x, posbk.y);				
				pos_time = t;
				newpos = true;
				update_seen_time = t;
			}
		}
	}
}


void MySelf::estimate_pos(Time time){
	if (pos_time < time){
		gpos.Setdata(Pos(pos_time) + SummarizeVelPos(pos_time, time), time);
		pos_time = time;
	}
}

Vector MySelf::SummarizeVelPos(Time prev, Time last){
	Vector moved_pos = 0;
	for(Time time = prev; time < last; time ++){
		moved_pos += Vel(time + 1) / speed_decay;
	}
	return moved_pos;
}

Vector MySelf::SummarizePosGap(Time prev, Time last){
	Vector moved_pos = 0;
	for(Time time = prev + 1; time <= last; time ++){
		moved_pos += pos_gap.ChkData(time);
	}
	return moved_pos;
}

bool MySelf::SelfColl(Time prevtime,Time lastime ){
	for(Time t=prevtime; t<=lastime;t++){
		if(selfcoll.ChkData(t)) return true;
	}
	return false;
}


Vector MySelf::PredictPos(float dash_pow, int cycles){
	Vector predictpos = pos;
	Vector vel = global_vel;
	int i;
	for(i = 0; i < cycles; i ++){
		vel += Polar2Vector(dash_pow, Self.bodyfacing) * Skill::action.Dashrate();
		if (vel.mod() > max_speed){
			vel = vel / vel.mod() * max_speed;
		}
		predictpos += vel;
		vel *= speed_decay;
	}
	return predictpos;
}

Vector MySelf::PredictPos(int cycles){
	return MobileObject::PredictPos(cycles);
}

Vector MySelf::PredictPos(AngleDeg facing, float dash_pow, int cycles){
	Vector predictpos = pos;
	Vector vel = global_vel;
	int i;
	for(i = 0; i < cycles; i ++){
		vel += Polar2Vector(dash_pow, facing) * Skill::action.Dashrate();
		if (vel.mod() > ServerParam::player_speed_max){
			vel = vel / vel.mod() * max_speed;
		}
		predictpos += vel;
		vel *= speed_decay;
	}
	return predictpos;
}

Vector MySelf::PredictPosWithTurnDash(AngleDeg angle, float dash_pow){
	Vector predictpos = Self.PredictPos(2);
	predictpos += Polar2Vector(dash_pow * Skill::action.Dashrate(), angle);
	return predictpos;
}

AngleDeg MySelf::MyViewWidth(VIEWWIDTH view_width){
	switch(view_width){
	case VW_Normal : return 90;
	case VW_Narrow : return 45;
	case VW_Wide   : return 180;
	}
	return 0.0f;
}

AngleDeg MySelf::MyViewAngle(VIEWWIDTH view_width){
	return MyViewWidth(view_width) / 2.0f;
}

AngleDeg MySelf::MyViewWidth(VWTime time){
	VIEWWIDTH VWidth = (time == VWT_Current) ? sensory.ViewWidth : sensory.NextViewWidth;
	return MyViewWidth(VWidth);
}


AngleDeg MySelf::MyViewAngle(VWTime time){
	return MyViewWidth(time) / 2.0f;
}

bool MySelf::Is_Rec_Sight_Info(Time t){
	return sensory.rec_sight_info.ChkData(t);
}
