#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "client.h"
#include "sensory.h"
#include "stopwatch.h"
#include "utils.h"
#include "worldmodel.h"
#include "perceptron.h"
#include "skill.h"
#include "log.h"
#include "agent.h"
#include "netif.h"

using namespace World;
/*****************************  Sensory  *******************************/
void Sensory::Initialize(){
	latestsenseinfo = 0;
	LastSightTime = -1;
	LastSenseTime = -1;
	PrevSenseTime = -1;
	LastSoundTime = -1;
	LastClosestMarker = No_Marker;
	dir_time = -1;
	int i;
	for(i = 0; i < ClientParam::num_divisions; i ++){
		FieldInfo.dir_conf[i] = 0.0f;
	}
	desire_viewidth = VW_Normal;
	desire_quality = VQ_High;
}

void Sensory::StopClock(bool value){
	situation.ClockStopped = value;

/*	Stop log when colck is stopped and restore when open */
	Log::rec.PauseLog(value);
	Log::recmsg.PauseLog(value);
}

/*****************************  Parse  **********************************/
/*	This Parse Module is transgrafted from CMU99 publication code.	We owe special
thanks to CMU's great effort	*/
void Sensory::Parse(char *SensoryInfo)
{
	SenseType sense_type;
	Time time;
	
	PrevInterruptTime = LatestTime;
	
	switch ( SensoryInfo[3] ){
	case 'e': sense_type = See_Msg;   break; /* see   */
	case 'n': sense_type = Sense_Msg; break; /* sense */
	case 'r': 
		switch(SensoryInfo[1]){
		case 's': sense_type = SP_Msg;    break; /* server params */
		default: sense_type = No_Msg;			/* maybe (error,warning...) */
		}
		break;	
		case 'a':								 /* hear, player params or player types */
			switch( SensoryInfo[1]){
			case 'c': sense_type = CPT_Msg;	break;	/*change player type */
			case 'h': sense_type = Hear_Msg;  break; /* hear */
			case 'p': /* player params or player types */
				switch(SensoryInfo[8]){
				case 'p': sense_type = PP_Msg; break; /* palyer params */
				case 't': sense_type = PT_Msg; break; /* player types */
				default : sense_type = No_Msg;
				}
				break;
				default: sense_type = No_Msg;
			}
			break;
			case 'o':
				sense_type = SC_Msg; 
				break;
			case 'l':
				sense_type = FI_Msg; 
				break;
			default:  sense_type = No_Msg;
	}
	switch ( sense_type ){		
		/* sense */
	case Sense_Msg: 
		// extract time info
		time = get_int(&SensoryInfo);
		if(LatestTime == time){
			StopClock(true);
		}else if(situation.ClockStopped){
			StopClock(false);
		}
		LatestTime = time;
		
		stopwatch.IdentifyCachetime(SW_CycleStart, time);//when the cycle starts

		Parse_Sense(time, SensoryInfo);
		LastSenseTime = time;
		NewSenseInfo = true;
		
		break;
		/* see */
	case See_Msg:
#ifdef _Up_FState
		return;
#endif		
		stopwatch.IdentifyCachetime(SW_Recv_Visualinfo, LatestTime);
		
		Make_ViewModeDecision();
		// extract time info
		time = get_int(&SensoryInfo);
		PrevSightTime = LastSightTime;
		LastSightTime = time;
		
		/* Don't parse a second sight from the a cycle prior to the latest one*/
		if (time <= PrevSightTime) break;/*&& !situation.ClockStopped*///abort sight info when clock stopped
		
		Parse_Sight(time, SensoryInfo);
		
		if(TwiceSight) Reset();//too much may result int out of array limit
		
		if(NewSight) TwiceSight = true;
		else NewSight = true;
		break;
		
	case FI_Msg:
		time = get_int(&SensoryInfo);
		PrevSightTime = LastSightTime;
		LastSightTime = time;
		Parse_Fullstate(time, SensoryInfo);
		stopwatch.IdentifyCachetime(SW_Recv_Fullstateinfo, sensory.LatestTime);
		break;
		
		/* hear */
	case Hear_Msg:  
		// extract time info
		time = get_int(&SensoryInfo);
		nonempty_msg = false;
		Parse_Sound(time, SensoryInfo); 
		LastSoundTime = time;
		NewSound = true;
		break;
	case SP_Msg:
		Parse_ServerParams(SensoryInfo + 13);
		break;
	case PP_Msg:
		Parse_PlayerParams(SensoryInfo +13);
		break;
	case PT_Msg:
		Parse_PlayerTypes(SensoryInfo + 12);
		break;
		
	case CPT_Msg:
		//Format "change_player_type 3 1"
		Parse_PlayerTypeChange(SensoryInfo + 20);
		break;
	case SC_Msg://scorez
		break;
	}
	
	LastSenseType = sense_type;
}

void Sensory::Parse_Sense(Time time, char *SenseInfo)
{
	get_word(&SenseInfo);  
	SenseInfo += 10;       /* "view_mode " */
	switch ( SenseInfo[0] )
	{	
	case 'h': ViewQuality = VQ_High; break;  /* high */
	case 'l': ViewQuality = VQ_Low;  break;  /* low  */
	default:  my_error("Unknown view quality");
	}
	
	LastViewWidth = ViewWidth;
	get_next_word(&SenseInfo);
	
	switch ( SenseInfo[1] )
	{
	case 'o': ViewWidth = VW_Normal; break;  /* normal */
	case 'a': ViewWidth = VW_Narrow; break;  /* narrow */
	case 'i': ViewWidth = VW_Wide;   break;  /* wide   */
	default:  my_error("Unknown view quality");
	}
	
	float stamina = get_float(&SenseInfo); /* stamina */
	float effort  = get_float(&SenseInfo); /* effort */
	float speed   = get_float(&SenseInfo); /* speed */
	float rel_spd_angle = get_float(&SenseInfo); /* rel_spd_angle */
	float head_angle = get_float(&SenseInfo); /* head angle */
	
	int kicks  =   get_int(&SenseInfo); /* kicks */
	int dashes =   get_int(&SenseInfo); /* dashes */
	int turns  =   get_int(&SenseInfo); /* turns */
	int says   =   get_int(&SenseInfo); /* says */
	int turn_necks   =   get_int(&SenseInfo); /* turn_necks */

	int catchs	=   get_int(&SenseInfo);
	int moves	=   get_int(&SenseInfo);
	int chang_view =   get_int(&SenseInfo);
		SetMySensedInfo(ViewWidth, ViewQuality,stamina,effort,speed,rel_spd_angle,head_angle,kicks,dashes,turns,says,turn_necks,time);
	
	if (ServerParam::version >=8.0f){
		if ( LastSenseTime == time && !situation.ClockStopped)	return;
	
		int ArmMoveable = get_int(&SenseInfo);
		int ArmExpires = get_int(&SenseInfo);
		Vector ArmTarget(get_float(&SenseInfo),get_float(&SenseInfo));
		int Arms = get_int(&SenseInfo);
		int Focus = -1;
		SenseInfo +=18;

		char side;
		switch(*SenseInfo){
		case 'n': 
			Focus = -1; 
			break;
		default:
			side = *SenseInfo;
			Focus = get_int(&SenseInfo);
			if(side != situation.MySide){
				Focus += SP_team_size;
			}
			break;
		}

		int Focuses = get_int(&SenseInfo);
		int TackleExpire = get_int(&SenseInfo);
		int Tackles = get_int(&SenseInfo);
		sensebodyinfo[latestsenseinfo].setvalues(
			ArmMoveable, ArmExpires, ArmTarget,	Arms,
			Focus,	Focuses, TackleExpire, Tackles);
	}
}

void Sensory::SetMyNumber(UNum no){
	Agent::MyNumber = no;

	Log::rec.SetID(no);
	Log::recmsg.SetID(no);
}

bool Sensory::Parse_initialize_message(char* pstr){
	char mode[100];	
	if (!(strncmp(pstr, "(init", 5)) ) {
		/* It's an init msg */
		UNum no;
		sscanf(pstr,"(init %c %d %[^)]",&situation.MySide, &no, mode);
		SetMyNumber(no);
		if (situation.MySide == 'r'){
			situation.TheirSide = 'l';
			MyGoal = Goal_R;
			TheirGoal = Goal_L;
		}
		else{
			situation.TheirSide = 'r';
			MyGoal = Goal_L;
			TheirGoal = Goal_R;
		}
	}
	else if ( !(strncmp(pstr,"(reconnect",5)) ) {
		/* It's a reconnect msg */
		sscanf(pstr,"(reconnect %c %[^)]",&situation.MySide, mode);
		SetMyNumber(ServerParam::IP_reconnect);
		if (situation.MySide == 'r'){
			situation.TheirSide = 'l';
			MyGoal = Goal_R;
			TheirGoal = Goal_L;
		}
		else{
			situation.TheirSide = 'r';
			MyGoal = Goal_L;
			TheirGoal = Goal_R;
		}
	}
	else {
		return false;
	}
	
	if ( ClientParam::goalie && FieldParam::FP_goalie_number != Agent::MyNumber )
		my_error("goalie number inconsistent with me being goalie");
	
	if ( !ClientParam::goalie && FieldParam::FP_goalie_number == Agent::MyNumber )
		my_error("I should be the goalie");
	
	if ( mode[0] == 'b' ){ /* Before_kick_off */ 
		situation.HearMode(PM_Before_Kick_Off);
		if ( situation.MySide == 'l' )
			situation.SetKickOffMode(KickOff_Mine);
		else 
			situation.SetKickOffMode(KickOff_Theirs);
	}
	else                  /* Act as if the game's in progress */
		situation.HearMode(PM_Play_On);
	
	return true;
}

/********************Set Sensed Infomation******************************************/

void Sensory::SetMySensedInfo(VIEWWIDTH viewidth, VIEWQUALITY viewqual, float stamina,float effort,float speed,AngleDeg rel_spd_angle,
							  AngleDeg head_angle,int kicks,int dashes,int turns,int says,int turn_necks,Time time){
	if ( LastSenseTime == time && !situation.ClockStopped)
		return;
	
	PrevSenseTime = LastSenseTime;
	LastSenseTime = time;
	
	latestsenseinfo = !latestsenseinfo;
	sensebodyinfo[latestsenseinfo].setvalues(viewidth, viewqual, stamina,effort,speed,rel_spd_angle,head_angle,kicks,dashes,
		turns,says,turn_necks,time);
}

SenseBodyInfo& Sensory::GetSenseBodyInfo(Time time){
	return (time == LastSenseTime) ? sensebodyinfo[latestsenseinfo] : sensebodyinfo[!latestsenseinfo];
}


/*******************************************************************************/

void Sensory::Update(){
	//Update Time
	situation.UpdateTime(LatestTime);
	Log::rec.SetTime(LatestTime);
	Log::recmsg.SetTime(LatestTime);
	
	if ( NewSight && LastSightTime < situation.CurrentTime - 1 && !situation.ClockStopped){
    /* Special case--ignore sights from just before change to play_on_mode
	if they seem to be 2 cycles ago.  Better would be to adjust the sight
		time, but then sight times of all other objects also need to be adjusted */
		/* process only recent cycles */
		//cancel by yjy
		//NewSight = false; 
		//rec_sight_info.Data(LastSightTime) = false;
		if ( LastSightTime < situation.CurrentTime - 2 )
			my_error("Last sight shouldn't be so out of date");
	}
	
	Skill::action.ActionsInvalidation();			// is the last cycle actions taking effect?
	Self.update(situation.CurrentTime);		// update self
	Update_others(situation.CurrentTime);	
	
	Update_dirconf(situation.CurrentTime, NewSight);
	Skill::action.PreProcess();
	
	if (situation.ClockStopped)
		situation.ClockStoppedTime ++;
	Reset();

	situation.SightTime = sensory.LastSightTime;

	situation.UpdateMode();

	visualsystem.PredictNextSightInf(); 
}

void Sensory::Update_dirconf(Time time, bool NewSight){
	int i;
#ifdef _Up_FState
	for(i = 0; i < ClientParam::num_divisions; i ++){
		FieldInfo.dir_conf[i] = ClientParam::conf_max;
	}
	dir_time = time;
	return;
#endif
	for(i = 0; i < ClientParam::num_divisions; i ++){
		FieldInfo.dir_conf[i] *= Exp(ClientParam::conf_decay, (float)(time - dir_time));
	}
	if (NewSight){
		for(i = 0; i < ClientParam::num_divisions; i ++){
			if (fabs(NormalizeAngle((i + 0.5f) * ClientParam::division - Self.Headfacing(LastSightTime))) < 0.8f * Self.MyViewAngle()){
				FieldInfo.dir_conf[i] = ClientParam::conf_max;
				if (LastSightTime != time)
					FieldInfo.dir_conf[i] *= Exp(ClientParam::conf_decay, (float)(time -LastSightTime));
			}
		}
	}
	dir_time = time;
	FieldInfo.Update_dirdanger(situation.CurrentTime);
}

void Sensory::Reset(){
	NewSight = false;
	NewSenseInfo = false;
	NewSound = false;
	TwiceSight = false;
}



bool Sensory::AppointUnkPlayer(Player* mp, VisualUnkPlayer* p){
	if(mp == NULL || p == NULL) return false;
	if(mp->IsSamePlayer(*p)){
		if(mp->seen_time() < p->t){
			mp->set_polar(p->seen_rel_angle(), p->seen_distance(), p->t);
			if(p->Is_side_known){ 
				mp->set_arm_tackle(p->armdir, p->tackling, p->t);
			}
			mp->update(p->t);
		}
		DoLog(LOG_UNKPLAYER,"Findnum %d pos(%.2f,%.2f)", mp->InsideNO, mp->pos.x, mp->pos.y);
		p->Recognize();
		return true;
	}else{
		return false;
	}
}

void Sensory::IdentifyUnkPlayers(Time t){
	int i, j;
	float dist, min_dist;
	Vector pos;
	VisualUnkPlayer* p;
	Player* matchp;
	Player* matched_player;

	int num_unknown_theirplayers = Their_Unk_Players.Data(t).Num_datas();
	//Note: when a unk players is identified successfully, it's removed from the unk array immediately
	// so descent order must be kept to make the remove action simplest.
	for(i = num_unknown_theirplayers-1; i >= 0; i--){
		//descent order to reduce update when twice sight
		p = & Their_Unk_Players.Data(t).Data(i);
		matched_player = NULL;
		min_dist = ClientParam::too_different_distance;

		for(j = 0; j < SP_team_size; j ++){
			matchp = & TheirTeam[j];
			if(matchp->IsSeen(t)) continue;
			if(matchp->pos_conf > ClientParam::identify_min_conf || matchp->Is_goalie){
				dist = matchp->pos.disaeolus(p->pos);
				if (dist < min_dist){
					matched_player = matchp;
					min_dist = dist;
				}
			}
		}
		if(AppointUnkPlayer(matched_player, p)) continue;

		matched_player = NULL;
		for(j=0; j<UnknownTheirPlayerList.oldnumplayers; j++){
			matchp = UnknownTheirPlayerList.List[j];
			if(matchp->IsSeen(t)) continue;
			if(matchp->pos_conf > ClientParam::identify_min_conf){
				dist = matchp->pos.disaeolus(p->pos);
				if (dist < min_dist){
					matched_player = matchp;
					min_dist = dist;
				}
			}
		}
		if(AppointUnkPlayer(matched_player, p)) continue;
		UnknownTheirPlayerList.EnList(*p);
	}

	int num_unknown_myplayers = My_Unk_Players.Data(t).Num_datas();
	for(i = num_unknown_myplayers-1; i >= 0; i--){
		p = & My_Unk_Players.Data(t).Data(i);
		matched_player = NULL;
		min_dist = ClientParam::too_different_distance;
		
		for(j = 0; j < SP_team_size; j ++){
			matchp = & MyTeam[j];
			if(matchp->InsideNO == Agent::MyNumber) continue;
			if(matchp->IsSeen(t)) continue;
			if(matchp->pos_conf > ClientParam::identify_min_conf){
				dist = matchp->pos.disaeolus(p->pos);
				if (dist < min_dist){
					matched_player = matchp;
					min_dist = dist;
				}
			}
		}
		if(AppointUnkPlayer(matched_player, p)) continue;

		matched_player = NULL;
		for(j=0; j<UnknownMyPlayerList.oldnumplayers; j++){
			matchp = UnknownMyPlayerList.List[j];
			if(matchp->IsSeen(t)) continue;
			if(matchp->pos_conf > ClientParam::identify_min_conf){
				dist = matchp->pos.disaeolus(p->pos);
				if (dist < min_dist){
					matched_player = matchp;
					min_dist = dist;
				}
			}
		}
		if(AppointUnkPlayer(matched_player, p)) continue;
		UnknownMyPlayerList.EnList(*p);
	}

	/*	Feeled Player Update	*/
	int num_teamlessplayers = Teamless_Players.Data(t).Num_datas();
	for(i = num_teamlessplayers-1; i >= 0; i--){		
		p = & Teamless_Players.Data(t).Data(i);
		matched_player = NULL;
		min_dist = ClientParam::too_different_distance;
		//ֻ֪Χڵteamless player
		if (p->seen_distance() > ServerParam::feel_distance) continue;
		
		for(j = 0; j < 2*SP_team_size; j ++){
			matchp = &GetPlayer(j+1);
			if(matchp->InsideNO == Agent::MyNumber) continue;
			if(matchp->IsSeen(t)) continue;
			if(matchp->pos_conf > ClientParam::identify_min_conf){
				dist = matchp->pos.disaeolus(p->pos);
				if (dist < min_dist){
					matched_player = matchp;
					min_dist = dist;
				}
			}
		}
		if(AppointUnkPlayer(matched_player, p)) continue;
		
		min_dist = ClientParam::too_different_distance;
		for(j = 0; j < UnknownTheirPlayerList.oldnumplayers; j++){
			matchp = UnknownTheirPlayerList.List[j];
			if(matchp->IsSeen(t)) continue;
			if(matchp->pos_conf > ClientParam::identify_min_conf){
				dist = matchp->pos.disaeolus(p->pos);
				if (dist < min_dist){
					matched_player = matchp;
					min_dist = dist;
				}
			}
		}
		for(j = 0; j < UnknownMyPlayerList.oldnumplayers; j++){
			matchp = UnknownMyPlayerList.List[j];
			if(matchp->IsSeen(t)) continue;
			if(matchp->pos_conf > ClientParam::identify_min_conf){
				dist = matchp->pos.disaeolus(p->pos);
				if (dist < min_dist){
					matched_player = matchp;
					min_dist = dist;
				}
			}
		}		
		if(AppointUnkPlayer(matched_player, p)) continue;
		
		matched_player = NULL;
		for(j = 0; j < TeamlessPlayerList.oldnumplayers; j++){
			matchp = TeamlessPlayerList.List[j];
			if(matchp->IsSeen(t)) continue;
			if(matchp->pos_conf > ClientParam::identify_min_conf){
				dist = matchp->pos.disaeolus(p->pos);
				if (dist < min_dist){
					matched_player = matchp;
					min_dist = dist;
				}
			}
		}
		
		if(AppointUnkPlayer(matched_player, p)) continue;
		TeamlessPlayerList.EnList(*p);
	}
	//special consideration, find their goalie from teamless player
	if(FieldInfo.IsValidtheirgoalie() && num_teamlessplayers > 0){
		matchp = &TheirPlayer(FieldInfo.theirgoalie);
		p = NULL;
		VisualUnkPlayer* tmp;
		min_dist = 2 * ClientParam::too_different_distance;
		if(!matchp->IsSeen(t) && matchp->IsValidObject()){
			for(i = num_teamlessplayers-1; i >= 0; i--){
				tmp = & Teamless_Players.Data(t).Data(i);
				if(tmp->recognized) continue; 
				if(!PitchInfo.their_penaltyarea.IsWithin(tmp->pos)) continue;
				dist = matchp->pos.disaeolus(tmp->pos);
				if (dist < min_dist){
					p = tmp;
					min_dist = dist;
				}
			}
			if(AppointUnkPlayer(matchp, p)){
				DoLog("find their goalie %.0f, %.0f", p->pos.x, p->pos.y);
			}
		}
	}
	UnknownMyPlayerList.RefreshNumPlayers();
	UnknownTheirPlayerList.RefreshNumPlayers(); 
	TeamlessPlayerList.RefreshNumPlayers();
}

void Sensory::Update_others(Time time){
	if(visualsystem.Vinfo_coming && !sensory.NewSight){
		DoLog(LOG_VDEC,"coming visual haven't come");
	}
	if(auditorysystem.rec_hear_info.IsDataKnown(time-1)){ 
		DoLog(LOG_HEAR, "speaker%d time%d", auditorysystem.rec_hear_info.Data(time-1), time-1);
	}
	ball.Prior_update();
	ball.update(time);// ball info should be updated prior to the other objs

	int i;
	for(i = 0; i < SP_team_size; i ++){
		MyTeam[i].Prior_update(); 
		MyTeam[i].update(time);
	}
	
	for(i = 0; i < SP_team_size; i ++){
		TheirTeam[i].Prior_update(); 
		TheirTeam[i].update(time);
	}

	UnknownTheirPlayerList.Prior_update();
	UnknownMyPlayerList.Prior_update();
	TeamlessPlayerList.Prior_update();

	updated_sight.Reset(); 
	//Unknown Player Identification
	for(Time t = updated_sight_time+1; t<=LastSightTime; t++){
		if(!rec_sight_info.ChkData(t)) continue;
		UpdateUnkPlayers(t);
		IdentifyUnkPlayers(t);

		updated_sight_time = t;
		Self.updated_sight_time=updated_sight_time;
		updated_sight.Add(t); 
	}

	UnknownTheirPlayerList.update(time);
	UnknownMyPlayerList.update(time);
	TeamlessPlayerList.update(time);
	
	/***********************update end********************/
	ball.update_end(); 
	Self.update_end(); 
	for(i = 0; i < SP_team_size; i ++){
		MyTeam[i].update_end();
	}
	for(i = 0; i < SP_team_size; i ++){
		TheirTeam[i].update_end();
	}
	
	UnknownTheirPlayerList.update_end();
	UnknownMyPlayerList.update_end();
	TeamlessPlayerList.update_end();

#ifdef _LOG
	UnknownMyPlayerList.LogList();
	UnknownTheirPlayerList.LogList();
	TeamlessPlayerList.LogList();
#endif
}

//Interface for decision-making module to call
//desire configuration used as reference for switching view modes
void Sensory::SetDesireView(VIEWWIDTH width,VIEWQUALITY quality){
	desire_viewidth = width;
	desire_quality = quality;
	DoRecord("desire width%d, quality%d", width, quality);
}

//called at the time of receiving sight info
//to make sure no duplicate sight info received
void Sensory::Make_ViewModeDecision(){
	VIEWWIDTH width = desire_viewidth;
	VIEWQUALITY quality = desire_quality;
	
	float elapsed_time = stopwatch.MsBetween(SW_CycleStart, LatestTime, SW_Recv_Visualinfo, LatestTime);
		
	bool chv = false;
	if(width == VW_Narrow){	
		bool work = false;
		//´ӾϢʱ, -2ʱӵľֵ
		DoRecord(LOG_SEE,"elapsed_time %d",elapsed_time);
		if(elapsed_time + SightInterval(width) - 2 >= ServerParam::sense_body_step){
			if(ViewWidth != width){
				change_view(width, quality);
				chv = true;
			}
		}else{
			if(ViewWidth == width || ViewWidth == VW_Wide){
				change_view(VW_Normal, quality);
				chv = true;
			}
			work = true;
		}
		DoRecord(LOG_STOPWATCH, "%d narrow %.1f, work%d, chv%d", LatestTime, elapsed_time, work?1:0, chv?1:0);
	}else if(width != ViewWidth){
		change_view(width, quality);
		chv = true;
	}
	DoRecord(LOG_SEE,"ChangeView(%d) %d",chv,width);

	if(!situation.ClockStopped && !chv){
		float nextsight_time;
		//note: if no change view, the nextviewwidth is just my current viewwidth.
		nextsight_time = elapsed_time + SightInterval(sensory.NextViewWidth);
		Self.NextViewCycle = LatestTime + nextsight_time / ServerParam::simulator_step;
	}
}

//physical command to be called in info proc instead of ctrl proc.
void Sensory::change_view(VIEWWIDTH width, VIEWQUALITY quality){
	char cmd_viewbuf[32];
	char qual_string[8];
	char width_string[8];

	switch (quality){
	case VQ_High: sprintf(qual_string,"high"); break;
	case VQ_Low:  sprintf(qual_string,"low");  break;
	}
	switch (width){
	case VW_Narrow: sprintf(width_string,"narrow"); break;
	case VW_Normal: sprintf(width_string,"normal"); break;
	case VW_Wide:   sprintf(width_string,"wide");   break;
	} 

	sprintf(cmd_viewbuf, "(change_view %s %s)", width_string, qual_string);
	sensory.NextViewWidth = width;
	Network::send_message(cmd_viewbuf, Network::sock);

#ifndef _OffClient
	stopwatch.TimeEvent(SW_Event_ChangeView);
	
#ifdef _LOGMAT
	TIME_T t = stopwatch.GetRawData_Event(SW_Event_ChangeView);	
#ifdef WIN32
	LogMat(LM_Ass, "%d %d %I64d %d %d", Ass_Chview, SW_Event_ChangeView, stopwatch.FormatTime(t), width, quality);
#else // _LINUX
	LogMat(LM_Ass, "%d %d %lld %d %d", Ass_Chview, SW_Event_ChangeView, stopwatch.FormatTime(t), width, quality);
#endif
#endif

	if(!situation.ClockStopped){
		float nextsight_time;
		//note: if no change view, the nextviewwidth is just my current viewwidth.
		nextsight_time = SightInterval(width) - stopwatch.MsFromEvent(SW_Event_ChangeView, SW_CycleStart, LatestTime);

		Self.NextViewCycle=LatestTime + nextsight_time / ServerParam::simulator_step;
	}
#endif
}

bool Sensory::HasFullState(){
#ifdef _Up_FState
	return true;
#else
	return false;
#endif
}

