
#include "player.h"
#include "unknownplayer.h"
#include "worldmodel.h"
#include "perceptron.h"
#include "log.h"

/*	UnknownPlayer	*/
UnknownPlayerNode::UnknownPlayerNode(){
	next = NULL;
}

/*	UnknownPlayerList	*/
UnknownPlayerList::UnknownPlayerList(){
	Head.next = NULL;
	Tail = &Head;
	numplayers = 0;
}

UnknownPlayerList::~UnknownPlayerList(){
	UnknownPlayerNode* tmp;
	while(Head.next != NULL){
		tmp = Head.next->next;
		delete Head.next;
		Head.next = tmp;
	}
}

void UnknownPlayerList::EnList(VisualUnkPlayer& player){
	if(numplayers >= SP_team_size){
		DoLog(LOG_UNKPLAYER,"num_unk_players reach limts");
		return;
	}

	Tail->next = new UnknownPlayerNode;
	Tail->next->unknownplayer.set_polar(player.seen_rel_angle(), player.seen_distance(), player.t);
	Tail->next->unknownplayer.set_side(side);
	Tail->next->unknownplayer.update(situation.CurrentTime);
	Tail = Tail->next;
	Tail->next = NULL;
	ArraySet(&(Tail->unknownplayer));
}

bool UnknownPlayerList::Clearable(Player& player){
	if(situation.ClockStopped) return true;
	if(player.Updated_pos()) return false;
	if(player.pos_conf < 0.8f) return true;

	if(!sensory.NewSight) return false;
	if(player.inview(sensory.LastSightTime) || (sensory.TwiceSight && player.inview(sensory.PrevSightTime)) || player.distance < ServerParam::feel_distance){
		if(IsUnknownside()) return true;
		if(player.distance < ServerParam::team_far_length) return true;
	}

	int i;
	float max_deviation = ClientParam::different_distance;
	if(IsMyside() || IsUnknownside()){
		for(i=0; i<SP_team_size; i++){
			if(!MyTeam[i].Updated_pos()) continue;
			if(player.pos.disaeolus(MyTeam[i].pos) < max_deviation) return true;
		}
	}
	if(IsOppside() || IsUnknownside()){
		for(i=0; i<SP_team_size; i++){
			if(!TheirTeam[i].Updated_pos()) continue;
			if(player.pos.disaeolus(TheirTeam[i].pos) < max_deviation) return true;
		}
	}
	return false;
}

void UnknownPlayerList::ArraySet(Player* player){
	List[numplayers ++] = player;
}

void UnknownPlayerList::update_end(){
	int i;
	for(i=0; i< numplayers; i++){
		List[i]->update_end();
	}
} 

void UnknownPlayerList::Prior_update(){
	int i;
	for(i=0; i< numplayers; i++){
		List[i]->Prior_update();
	}
}

void UnknownPlayerList::update(Time time){
	UnknownPlayerNode* p = &Head, *tmp;
	numplayers = 0;	
	while(p->next != NULL){		
		if (Clearable(p->next->unknownplayer)){
			tmp = p->next->next;
			delete p->next;
			p->next = tmp;
		}else{
			p = p->next;
			p->unknownplayer.update(time);
			ArraySet(&(p->unknownplayer));
		}
	}
	Tail = p;
	RefreshNumPlayers();
}

void UnknownPlayerList::LogList(char* msg){
	char logbuf[1024] = "";
	UnknownPlayerNode* p = &Head;	
	int logcount=0;
	if(msg != NULL)
		logcount += sprintf(logbuf, "%s", msg);
	else{
		if(IsMyside()){
			logcount += sprintf(logbuf,"Teammate %d", numplayers);
		}else if(IsOppside()){
			logcount += sprintf(logbuf,"Opp %d", numplayers);
		}else{
			logcount += sprintf(logbuf,"Unknown %d", numplayers);
		}
	}
	while(p->next != NULL){
		p = p->next;
		logcount += sprintf(& logbuf[logcount]," pos(%.2f,%.2f)",p->unknownplayer.pos.x, p->unknownplayer.pos.y);
	}
	DoLog(LOG_UNKPLAYER, logbuf);
}

int UnknownPlayerList::NumUnknownPlayers(){
	return numplayers;
}

void UnknownPlayerList::Set_IsControlball_FromHeard(Time time){
	UnknownPlayerNode* p = &Head;
	while(p->next != NULL){
		p->next->unknownplayer.Set_Iscontrolball_Fromheard(time);
		p = p->next;
	}
}

